/* compound.c -- Compound object implementation.
 * Copyright (C) 2014 Wolfgang Jaehrling
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "compound.h"
#include "mem.h"
#include "symkey.h"
#include "num.h"
#include "eval.h"
#include "hash.h"

static PWG_GCC_ALWAYS_INLINE pwg_t
reference (pwg_t o, pwg_type_t t, natural_t n)
{
  return ((pwg_t *) mem_ref_obj (o, t))[n];
}

static PWG_GCC_ALWAYS_INLINE void
assign (pwg_t o, pwg_type_t t, natural_t n, pwg_t x)
{
  ((pwg_t *) mem_ref_obj (o, t))[n] = x;
}

#define DEF_MARKER(name, type, fields)		\
void						\
name (pwg_t arg)				\
{						\
  int i;					\
  for (i = 0; i < fields; i++)			\
    gc_mark (reference (arg, type, i));		\
}						\
extern void dummy ()

#define DEF_ACCESSOR(name, type, pos)		\
pwg_t						\
name (pwg_t arg)				\
{						\
  return reference (arg, type, pos);		\
}						\
extern void dummy (void)

#define DEF_SETTER(name, type, pos)		\
pwg_t						\
name (pwg_t arg, pwg_t x)			\
{						\
  assign (arg, type, pos, x);			\
  return PWG_NONE;				\
}						\
extern void dummy (void)

#define DEF_COPY(name, o, type, size)				\
pwg_t								\
name (pwg_t o)							\
{								\
  pwg_t fresh = gc_alloc (type, sizeof (pwg_t) * (size));	\
  for (integer_t i = 0; i < (size); i++)			\
    assign (fresh, type, i, pwg_copy (reference (o, type, i)));	\
  return fresh;							\
}								\
extern void dummy (void)



/* The characters stuff is implemented here because it is considered a
   part of the string implementation.  Now if *that* ain't a bad
   excuse...  */

bool
chr_p (pwg_t c)
{
  return PWG_BITEST (c, 3, 4);
}

integer_t
unchr (pwg_t c)
{
  if (!chr_p (c))
    err ("type error: accessing non-chr as chr");
  return pwg_get (c) >> 3;
}

pwg_t
chr (integer_t n)
{
  return pwg_new ((n << 3) | 3);
}

pwg_t
pwg_chr_to_sint (pwg_t c)
{
  return sint (unchr (c));
}

pwg_t
pwg_sint_to_chr (pwg_t i)
{
  return chr (unsint (i));
}

integer_t
chr_digit_to_int (pwg_t c)
{
  integer_t tmp = unchr (c) - '0';
  return (tmp >= 0 && tmp <= 9) ? tmp : -1;
}

pwg_t
pwg_digit_to_sint (pwg_t c)
{
  integer_t tmp = chr_digit_to_int (c);
  return (tmp == -1) ? PWG_FALSE : sint (tmp);
}

DEF_PRED (pwg_chr_p, o, chr_p (o));



pwg_t
pwg_cons (pwg_t car, pwg_t cdr)
{
  pwg_t obj = gc_alloc (TYPE_CONS, sizeof (pwg_t) * 2);

  pwg_set_car (obj, car);
  pwg_set_cdr (obj, cdr);
  return obj;
}

DEF_MARKER (cons_gc_mark, TYPE_CONS, 2);
DEF_COPY (pwg_cons_copy, o, TYPE_CONS, 2);
DEF_PRED (pwg_cons_p, o, mem_is_type_p (o, TYPE_CONS));
DEF_ACCESSOR (pwg_car, TYPE_CONS, 0);
DEF_ACCESSOR (pwg_cdr, TYPE_CONS, 1);
DEF_SETTER (pwg_set_car, TYPE_CONS, 0);
DEF_SETTER (pwg_set_cdr, TYPE_CONS, 1);

/* List functions needed in the core.  */

pwg_t
pwg_size (pwg_t lst)
{
  integer_t cnt = 0;

  LIST_FOR_EACH (lst, ptr)
    cnt++;
  return sint (cnt);
}

pwg_t
pwg_reverse (pwg_t lst)
{
  pwg_t r = PWG_NIL;

  LIST_FOR_EACH (lst, ptr)
    r = pwg_cons (pwg_car (ptr), r);
  return r;
}

pwg_t
pwg_reverse_x (pwg_t lst)
{
  pwg_t old, new = PWG_NIL;

  while (!eq_p (lst, PWG_NIL))
    {
      old = pwg_cdr (lst);
      pwg_set_cdr (lst, new);
      new = lst;
      lst = old;
    }
  return new;
}

pwg_t
pwg_shallow_copy (pwg_t lst)
{
  /* Slow but simple.  */
  return pwg_reverse_x (pwg_reverse (lst));
}



pwg_t
str_new (const char *s)
{
  size_t len = strlen (s);
  pwg_t o = gc_alloc (TYPE_STR, sizeof (pwg_t) * (len + 1));

  assign (o, TYPE_STR, 0, sint (len));
  for (integer_t i = 1; *s != '\0'; s++, i++)
    assign (o, TYPE_STR, i, chr (*s));
  return o;
}

pwg_t
pwg_language_version (void)
{
  return str_new (PWG_VERSION);
}

PWG_GCC_MALLOC char *
unstr (pwg_t str)
{
  /* FIXME: Update for Unicode.  */
  integer_t i;
  integer_t len = unsint (reference (str, TYPE_STR, 0));
  char *result = mem_alloc (len + 1, "ascii string from pwg string");

  for (i = 0; i < len; i++)
    result[i] = unchr (reference (str, TYPE_STR, i + 1));
  result[i] = '\0';
  return result;
}

bool
str_eql (pwg_t str, const char *cstr)
{
  return unbooled (pwg_str_eql_p (str_new (cstr), str));
}

DEF_COPY (pwg_str_copy, o, TYPE_STR, unsint (pwg_str_size (o)) + 1);
DEF_PRED (pwg_str_p, o, mem_is_type_p (o, TYPE_STR));
DEF_ACCESSOR (pwg_str_size, TYPE_STR, 0);

pwg_t
pwg_str_append (pwg_t s, pwg_t t)
{
  integer_t i;
  integer_t s_len = unsint (pwg_str_size (s));
  integer_t t_len = unsint (pwg_str_size (t));
  pwg_t o = gc_alloc (TYPE_STR, sizeof (pwg_t) * (s_len + t_len + 1));

  assign (o, TYPE_STR, 0, sint (s_len + t_len));
  for (i = 0; i != s_len; i++)
    assign (o, TYPE_STR, i + 1, pwg_str_ref (s, sint (i)));
  for (integer_t j = 0; j != t_len; i++, j++)
    assign (o, TYPE_STR, i + 1, pwg_str_ref (t, sint (j)));
  return o;

}

pwg_t
pwg_str_ref (pwg_t s, pwg_t p)
{
  integer_t pos = unsint (p);
  integer_t len = unsint (reference (s, TYPE_STR, 0));

  if (pos < 0 || pos >= len)
    err ("str index out of range");
  return reference (s, TYPE_STR, pos + 1);
}

/* LEN may be negative, meaning: to begin at the end.  */
pwg_t
pwg_str_slice (pwg_t s, pwg_t start, pwg_t len)
{
  if (unsint (len) < 0)
    {
      len = sint (-unsint (len));
      start = sint (unsint (start) - unsint (len));
    }

  pwg_t o = gc_alloc (TYPE_STR, sizeof (pwg_t) * (unsint (len) + 1));

  assign (o, TYPE_STR, 0, len);
  for (integer_t i = 1;
       i <= unsint (len);
       i++, start = sint (unsint (start) + 1))
    assign (o, TYPE_STR, i, pwg_str_ref (s, start));

  return o;  
}

pwg_t
pwg_str_eql_p (pwg_t s, pwg_t t)
{
  integer_t s_len = unsint (pwg_str_size (s));
  integer_t t_len = unsint (pwg_str_size (t));

  if (s_len != t_len)
    return PWG_FALSE;

  for (integer_t i = 0; i < s_len; i++)
    if (!eq_p (pwg_str_ref (s, sint (i)), pwg_str_ref (t, sint (i))))
      return PWG_FALSE;
  return PWG_TRUE;
}

// FIXME: kind of a strange function.only used internally, i think.
pwg_t
pwg_str_sym_eql_p (pwg_t s1, pwg_t s2)
{
  pwg_type_t t1 = mem_ref_type (s1);
  pwg_type_t t2 = mem_ref_type (s2);

  if ((t1 != TYPE_STR && t1 != TYPE_SYM) || (t2 != TYPE_STR && t2 != TYPE_SYM))
    err ("type error: expected sym and str");

  mem_set_type (s1, TYPE_STR);
  mem_set_type (s2, TYPE_STR);
  pwg_t result = pwg_str_eql_p (s1, s2);
  mem_set_type (s1, t1);
  mem_set_type (s2, t2);
  return result;
}

pwg_t
pwg_str_hash (pwg_t s)
{
  natural_t result, i;
  size_t len = unsint (pwg_str_size (s));

  if (len <= sizeof (pwg_t))
    for (i = result = 0; i < len; i++)
      result = (result << 8) + unchr (pwg_str_ref (s, sint (i)));
  else /* FIXME: use this else-part for all strs? */
    for (i = 0, result = len; i < (sizeof (pwg_t) + 2); i++)
      result = ((result << 6)
		+ unchr (pwg_str_ref (s, sint ((result + i) % len))));

  return sint (result & SINT_MAX);
}

/* Rename to pwg_str_new? */
pwg_t
pwg_list_to_str (pwg_t lst)
{
  pwg_t len = pwg_size (lst);
  pwg_t o = gc_alloc (TYPE_STR, sizeof (pwg_t) * (unsint (len) + 1));

  assign (o, TYPE_STR, 0, len);
  for (integer_t i = 1; !eq_p (lst, PWG_NIL); i++, lst = pwg_cdr (lst))
    {
      if (!chr_p (pwg_car (lst)))
	err ("type error: item in list is not a chr");
      assign (o, TYPE_STR, i, pwg_car (lst));
    }
  return o;
}

pwg_t
pwg_str_to_list (pwg_t s)
{
  pwg_t c = PWG_NIL;
  integer_t len = unsint (reference (s, TYPE_STR, 0));
  for (integer_t i = len; i >= 1; i--)
    c = pwg_cons (reference (s, TYPE_STR, i), c);

  return c;
}

pwg_t
pwg_str_to_sint (pwg_t s)
{
  integer_t result = 0;
  bool positive_p = true;
  bool found_num_p = false;
  pwg_t i;

  TOWARD (i, pwg_str_size (s))
    {
      integer_t tmp = chr_digit_to_int (pwg_str_ref (s, i));

      if (tmp == -1)
	{
	  if (eq_p (i, sint (0)) && eq_p (pwg_str_ref (s, i), chr ('-')))
	    {
	      positive_p = false;
	      continue;
	    }
	  else
	    return PWG_FALSE;
	}
      found_num_p = true;
      result *= 10;
      result += tmp;
    }
  return found_num_p ? (sint (positive_p ? result : -result)) : PWG_FALSE;
}

pwg_t
pwg_str_to_float (pwg_t s)
{
  /* Seems okay for Unicode.  */
  char *end, *str = unstr (s);

  errno = 0;
  double f = strtod (str, &end);
  bool success_p = (end == str + strlen (str));

  mem_free (str);
  if (errno != 0)
    err ("number cannot be represented");

  return success_p ? num_float (f) : PWG_FALSE;
}



// TODO: we should also be able to create an array with size N
pwg_t
pwg_ary_new (pwg_t lst)
{
  integer_t len = unsint (pwg_size (lst));
  pwg_t obj = gc_alloc (TYPE_ARY, sizeof (pwg_t) * (len + 1));

  assign (obj, TYPE_ARY, 0, sint (len));
  for (integer_t i = 1; i <= len; i++, lst = pwg_cdr (lst))
    assign (obj, TYPE_ARY, i, pwg_car (lst));

  return obj;
}

// FIXME: calls GC on the size entry (unneccessary)
DEF_MARKER (ary_gc_mark, TYPE_ARY, unsint (pwg_ary_size (arg)) + 1);
DEF_COPY (pwg_ary_copy, o, TYPE_ARY, unsint (pwg_ary_size (o)) + 1);
DEF_PRED (pwg_ary_p, o, mem_is_type_p (o, TYPE_ARY));
DEF_ACCESSOR (pwg_ary_size, TYPE_ARY, 0);

pwg_t
pwg_ary_item (pwg_t a, pwg_t n)
{
  integer_t pos = unsint (n);
  integer_t len = unsint (pwg_ary_size (a));

  if (pos < 0 || pos >= len)
    err ("ary index out of range");

  return reference (a, TYPE_ARY, pos + 1);
}

pwg_t
pwg_ary_set_item (pwg_t a, pwg_t n, pwg_t x)
{
  integer_t pos = unsint (n);
  integer_t len = unsint (pwg_ary_size (a));

  if (pos < 0 || pos >= len)
    err ("ary index out of range");

  assign (a, TYPE_ARY, unsint (n) + 1, x);
  return PWG_NONE;
}



pwg_t
pwg_closure_new (pwg_t obj, pwg_t env)
{
  pwg_t o = gc_alloc (TYPE_CLOSURE, sizeof (pwg_t) * 2);

  assign (o, TYPE_CLOSURE, 0, obj);
  assign (o, TYPE_CLOSURE, 1, env);
  return o;
}

// FIXME: is it correct to /not/ copy the hash?
// FIXME: isn't this shallow_copy?
static pwg_t
copy_env (pwg_t env)
{
  if (unbooled (pwg_cons_p (env)))
    return pwg_cons (pwg_copy (pwg_car (env)),
		     copy_env (pwg_cdr (env)));
  else
    return env;
}

pwg_t
pwg_closure_copy (pwg_t c)
{
  return pwg_closure_new (pwg_closure_data (c),
			  copy_env (pwg_closure_env (c)));
}

DEF_MARKER (closure_gc_mark, TYPE_CLOSURE, 2);
DEF_PRED (pwg_closure_p, o, mem_is_type_p (o, TYPE_CLOSURE));
DEF_ACCESSOR (pwg_closure_data, TYPE_CLOSURE, 0);
DEF_ACCESSOR (pwg_closure_env, TYPE_CLOSURE, 1);



pwg_t
pwg_proc_new (pwg_t code, pwg_t args, pwg_t extra_args, pwg_t re_p,
	      pwg_t open_space_p)
{
  pwg_t o = gc_alloc (TYPE_PROC, sizeof (pwg_t) * 5);

  assign (o, TYPE_PROC, 0, code);
  assign (o, TYPE_PROC, 1, args);
  assign (o, TYPE_PROC, 2, extra_args);
  assign (o, TYPE_PROC, 3, re_p);
  assign (o, TYPE_PROC, 4, open_space_p);
  return o;
}

DEF_MARKER (proc_gc_mark, TYPE_PROC, 3); /* Others are imms anyway. */
DEF_COPY (pwg_proc_copy, o, TYPE_PROC, 5);
DEF_PRED (pwg_proc_p, o, mem_is_type_p (o, TYPE_PROC));
DEF_ACCESSOR (pwg_proc_code, TYPE_PROC, 0);
DEF_ACCESSOR (pwg_proc_args, TYPE_PROC, 1);
DEF_ACCESSOR (pwg_proc_extra_args, TYPE_PROC, 2);
DEF_ACCESSOR (pwg_proc_re_p, TYPE_PROC, 3);
DEF_ACCESSOR (pwg_proc_open_space_p, TYPE_PROC, 4);



// FIXME: allow types with destructors.

static natural_t next_type = END_BUILTIN_TYPES;

pwg_type_t
type_new (void)
{
  return TYPE (next_type++);
}

pwg_t
pwg_type_new (void)
{
  return sint (type_new ());
}

pwg_t
pwg_wrap (pwg_t type, pwg_t obj)
{
  pwg_type_t t = unsint (type);
  if (t < TYPE (END_BASE_TYPES) || t >= TYPE (next_type))
    err ("type error: trying to wrap object into invalid type");

  pwg_t o = gc_alloc (t, sizeof (pwg_t));
  assign (o, t, 0, obj);
  return o;
}

pwg_t
pwg_unwrap (pwg_t obj)
{
  pwg_type_t type = mem_ref_type (obj);
  if (type < TYPE (END_BASE_TYPES) || type >= TYPE (next_type))
    err ("type error: unwrapping object with wrong type");
  return reference (obj, mem_ref_type (obj), 0);
}

#define DEF_WRAPPED(type, constructor)		\
pwg_t						\
constructor (pwg_t o)				\
{						\
  return pwg_wrap (sint (type), o);		\
} extern void dummy ()


DEF_WRAPPED (TYPE_QUOTOR, pwg_quotor_new);
DEF_WRAPPED (TYPE_SETTOR, pwg_settor_new);
DEF_WRAPPED (TYPE_PUSHOR, pwg_pushor_new);



pwg_t
pwg_copy (pwg_t o)
{
  if (PWG_IMMP (o))
    return o;
  pwg_type_t t = mem_ref_type (o);
  switch (t)
    {
    case TYPE_CONS: return pwg_cons_copy (o);
    case TYPE_SYM: return o;
    case TYPE_STR: return pwg_str_copy (o);
    case TYPE_ARY: return pwg_ary_copy (o);
    case TYPE_FLOAT: return num_float (num_unfloat (o));
    case TYPE_HASH: return pwg_hash_copy (o);
    case TYPE_PROC: return pwg_proc_copy (o);
    case TYPE_PSUB: return pwg_psub_copy (o);
    case TYPE_CLOSURE: return pwg_closure_copy (o);
    case TYPE_STREAM: err ("cannot create a copy of a stream");
    default:
      if (t < END_BUILTIN_TYPES || t >= next_type)
	err ("type error: object with invalid type?!");
      return pwg_wrap (sint (t), pwg_unwrap (o));
    }
  abort (); /* Not reached. */
  return PWG_VOID;
}



void
setup_compound (void)
{}

void
init_compound (void)
{
  def_core_psub ("digit->sint", pwg_digit_to_sint, 1);
  def_core_psub ("chr->sint", pwg_chr_to_sint, 1);
  def_core_psub ("sint->chr", pwg_sint_to_chr, 1);
  def_core_psub ("chr?", pwg_chr_p, 1);

  def_core_psub ("cons", pwg_cons, 2);
  def_core_psub ("cons-copy", pwg_cons_copy, 1);
  def_core_psub ("cons?", pwg_cons_p, 1);
  def_core_psub ("car", pwg_car, 1);
  def_core_psub ("cdr", pwg_cdr, 1);
  def_core_psub ("car=>", pwg_set_car, 2);
  def_core_psub ("cdr=>", pwg_set_cdr, 2);
  def_core_psub ("size", pwg_size, 1);
  def_core_psub ("reverse", pwg_reverse, 1);
  def_core_psub ("reverse!", pwg_reverse_x, 1);
  def_core_psub ("shallow-copy", pwg_shallow_copy, 1);

  def_core_psub ("ary-new", pwg_ary_new, 1);
  def_core_psub ("ary-copy", pwg_ary_copy, 1);
  def_core_psub ("ary?", pwg_ary_p, 1);
  def_core_psub ("ary-size", pwg_ary_size, 1);
  def_core_psub ("ary-item", pwg_ary_item, 2);
  def_core_psub ("ary-item=>", pwg_ary_set_item, 3);

  def_core_psub ("language-version", pwg_language_version, 0);
  def_core_psub ("str-copy", pwg_str_copy, 1);
  def_core_psub ("str?", pwg_str_p, 1);
  def_core_psub ("str-append", pwg_str_append, 2);
  def_core_psub ("str-ref", pwg_str_ref, 2);
  def_core_psub ("str-slice", pwg_str_slice, 3);
  def_core_psub ("str-size", pwg_str_size, 1);
  def_core_psub ("str-eql?", pwg_str_eql_p, 2);
  def_core_psub ("str-sym-eql?", pwg_str_sym_eql_p, 2);
  def_core_psub ("str-hash", pwg_str_hash, 1);
  def_core_psub ("list->str", pwg_list_to_str, 1);
  def_core_psub ("str->list", pwg_str_to_list, 1);
  def_core_psub ("str->sint", pwg_str_to_sint, 1);
  def_core_psub ("str->float", pwg_str_to_float, 1);

  /* FIXME: should we allow this?  */
  /* def_core_psub ("closure-new", pwg_closure_new, 2); */
  def_core_psub ("closure?", pwg_closure_p, 1);
  def_core_psub ("closure-copy", pwg_closure_copy, 1);
  def_core_psub ("closure-data", pwg_closure_data, 1);
  def_core_psub ("closure-env", pwg_closure_env, 1);

  def_core_psub ("proc-new", pwg_proc_new, 4);
  def_core_psub ("proc-copy", pwg_proc_copy, 1);
  def_core_psub ("proc?", pwg_proc_p, 1);
  def_core_psub ("proc-code", pwg_proc_code, 1);
  def_core_psub ("proc-args", pwg_proc_args, 1);
  def_core_psub ("proc-extra-args", pwg_proc_extra_args, 1);
  def_core_psub ("proc-re?", pwg_proc_re_p, 1);
  def_core_psub ("proc-open-space?", pwg_proc_open_space_p, 1);

  //FIXME: they cannot be printed yet, so don't allow to create them..
  //def_core_psub ("type-new", pwg_type_new, 0);
  //def_core_psub ("wrap", pwg_wrap, 2);
  def_core_psub ("unwrap", pwg_unwrap, 1);
  def_core_psub ("quotor-new", pwg_quotor_new, 1);
  def_core_psub ("settor-new", pwg_settor_new, 1);
  def_core_psub ("pushor-new", pwg_pushor_new, 1);

  def_core_psub ("copy", pwg_copy, 1);
}

void
finish_compound (void)
{}
