/* eval.c -- Evaluator implementation.
 * Copyright (C) 2014 Wolfgang Jaehrling
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdlib.h>
#include <ctype.h>
#include <stdio.h>

#include "eval.h"
#include "mem.h"
#include "compound.h"
#include "symkey.h"
#include "hash.h"
#include "stream.h"

#define DATA_STACK_INIT 32
#define DATA_STACK_FAST 1024
#define CALL_STACK_MAX 12000 /* Will crash otherwise.  FIXME: ???*/

// FIXME: GC? -> yes, at least re_sym needs to be GC'd
// but currently we don't collect syms anyway.
static pwg_t re_sym;
static pwg_t core_sym;
static pwg_t repl_sym;
static pwg_t used_modules_sym;



pwg_t
psub_new (psub_fn_t fn, integer_t args)
{
  pwg_t obj = gc_alloc (TYPE_PSUB,
			     sizeof (psub_fn_t) + sizeof (integer_t));
  void **data = mem_ref_obj (obj, TYPE_PSUB);

  *data = (void *) fn;
  *(data + 1) = (void *) args;
  return obj;
}

psub_fn_t
psub_fn (pwg_t obj)
{
  void **data = mem_ref_obj (obj, TYPE_PSUB);
  return (psub_fn_t) *data;
}

integer_t
psub_args (pwg_t obj)
{
  void **data = mem_ref_obj (obj, TYPE_PSUB);
  return (integer_t) *(data + 1);
}

DEF_PRED (pwg_psub_p, o, mem_is_type_p (o, TYPE_PSUB));

void
psub_call (pwg_t obj)
{
  pwg_t result;
  psub_fn_t fn = psub_fn (obj);
  integer_t args = psub_args (obj);

  switch (args)
    {
    case 0: result = ((pwg_t (*)(void)) fn) (); break;
    case 1: result = ((pwg_t (*)(pwg_t)) fn) (pop_data ()); break;
    case 2: {
      pwg_t a = pop_data (); pwg_t b = pop_data ();
      result = ((pwg_t (*)(pwg_t, pwg_t)) fn) (a, b); break;
    } case 3: {
      pwg_t a = pop_data (); pwg_t b = pop_data (); pwg_t c = pop_data ();
      result = ((pwg_t (*)(pwg_t, pwg_t, pwg_t)) fn) (a, b, c); break;
    } case 4: {
      pwg_t a = pop_data (); pwg_t b = pop_data ();
      pwg_t c = pop_data (); pwg_t d = pop_data ();
      result = ((pwg_t (*)(pwg_t,pwg_t,pwg_t,pwg_t)) fn) (a, b, c, d); break;
    } case 5: {
      pwg_t a = pop_data (); pwg_t b = pop_data (); pwg_t c = pop_data ();
      pwg_t d = pop_data (); pwg_t e = pop_data ();
      result = ((pwg_t (*)(pwg_t, pwg_t, pwg_t, pwg_t, pwg_t)) fn)
	(a, b, c, d, e);
      break;
    } case 6: {
      pwg_t a = pop_data (); pwg_t b = pop_data (); pwg_t c = pop_data ();
      pwg_t d = pop_data (); pwg_t e = pop_data (); pwg_t f = pop_data ();
      result = ((pwg_t (*)(pwg_t, pwg_t, pwg_t, pwg_t, pwg_t, pwg_t)) fn)
	(a, b, c, d, e, f);
      break;
    } default:
      err ("internal error: psub with too many (or less than zero?) args");
    }
  if (!eq_p (result, PWG_NONE))
    push_data (result);
}

pwg_t
pwg_psub_copy (pwg_t obj)
{
  return psub_new (psub_fn (obj), psub_args (obj));
}



#ifdef PWG_DEBUG
static bool tracing_p = false;

static void
trace_note (pwg_t obj, const char *desc)
{
  if (!tracing_p)
    return;
  pwg_stream_out (pwg_stream_stdout (), obj);
  printf (" # %s\n", desc);
}
#endif

static pwg_t modules;

void
mark_modules (void)
{
  gc_mark (modules);
}

/* How Environments work:
   An env. is an improper alist: the last cdr is a hash table that
   also has an entry that is a list of other environments. Lookup
   for reading processes these three places in order, while setting
   skips the last, so entries in other environments are read-only. */

/* Look up the SYM in ENV.  If there is no such entry, raise an
   error.  */
static pwg_t
lookup (pwg_t sym, pwg_t env)
{
  pwg_t result;

  for (; unbooled (pwg_cons_p (env)); env = pwg_cdr (env))
    if (unbooled (pwg_cons_p (pwg_car (env))))
      {
	if (eq_p (pwg_car (pwg_car (env)), sym))
	  return pwg_cdr (pwg_car (env));
      }
    else
      {
	result = pwg_hash_item (pwg_car (env), sym);
	if (!eq_p (result, PWG_NONE))
	  return result;
      }

  result = pwg_hash_item (env, sym);
  if (eq_p (result, PWG_NONE))
    {
      LIST_FOR_EACH (pwg_hash_item (env, used_modules_sym), ptr)
	{
	  pwg_t mod = pwg_hash_item (modules, pwg_car (ptr));

	  result = pwg_hash_item (mod, sym);
	  if (!eq_p (result, PWG_NONE))
	    return result;
	}
      exc_store (pwg_str_append (str_new ("symbol not found: "),
				 pwg_sym_to_str (sym)));
      exc_raise (EXC_ERR);
    }
  return result;
}

static void
lookup_and_set (pwg_t sym, pwg_t env, pwg_t val)
{
  for (; unbooled (pwg_cons_p (env)); env = pwg_cdr (env))
    if (unbooled (pwg_cons_p (pwg_car (env))))
      {
	if (eq_p (pwg_car (pwg_car (env)), sym))
	  {
	    pwg_set_cdr (pwg_car (env), val);
	    return;
	  }
      }
    else
      {
	pwg_hash_set_item (pwg_car (env), sym, val);
	return;
      }
  pwg_hash_set_item (env, sym, val);
}

void
def_core_psub_internal (const char *name, psub_fn_t fn, integer_t args)
{
  lookup_and_set (sym (name), pwg_core_module (), psub_new (fn, args));
}



#ifdef PWG_FAST
static pwg_t data_stack[DATA_STACK_FAST];
#else
static pwg_t *data_stack;
static integer_t data_stack_max;
#endif
static integer_t data_stack_size;

static pwg_t call_stack;

#ifndef PWG_FAST
static void
call_stack_out_var (pwg_t *dst, pwg_t name, pwg_t val)
{
  pwg_stream_out (*dst, name);
  pwg_stream_write (*dst, chr ('='));
  pwg_stream_out (*dst, val);
  pwg_stream_write (*dst, chr (' '));
}

void
call_stack_out (pwg_t dst, pwg_t stack)
{
  int frame = 0;

  /* Easier to read in reversed order. Speed is not an issue here anyway. */
  LIST_FOR_EACH (pwg_reverse (stack), ptr)
    {
      pwg_stream_out (dst, sint (frame));
      frame++;
      pwg_stream_write_str (dst, str_new (": "));

      if (unbooled (pwg_cons_p (pwg_car (ptr))))
	{
	  /* Code and environment.  */
	  pwg_stream_out (dst, pwg_car (pwg_car (ptr)));
	  pwg_stream_write_str (dst, str_new (" -> "));

	  for (pwg_t env_ptr = pwg_cdr (pwg_car (ptr));
	       unbooled (pwg_cons_p (env_ptr));
	       env_ptr = pwg_cdr (env_ptr))
	    if (unbooled (pwg_cons_p (pwg_car (env_ptr))))
	      call_stack_out_var (&dst, pwg_car (pwg_car (env_ptr)),
				  pwg_cdr (pwg_car (env_ptr)));
	    else
	      hash_iterate_pwg (pwg_car (env_ptr),
				(hash_iter_pwg_t) call_stack_out_var, &dst);
	}
      else
	pwg_stream_out (dst, pwg_car (ptr));
      pwg_stream_write (dst, chr ('\n'));
    }
}
#endif /* !PWG_FAST */

/* This gives a list of all items currently on the data stack.  */
pwg_t
pwg_data_stack (void)
{
  pwg_t result = PWG_NIL;

  for (integer_t i = 0; i < data_stack_size; i++)
    result = pwg_cons (data_stack[i], result);
  return result;
}

void
push_data (pwg_t obj)
{
#ifndef PWG_FAST
  if (data_stack_size == data_stack_max)
    {
      data_stack_max *= 2;
      pwg_t *stack = mem_alloc (sizeof (pwg_t) * data_stack_max,
				"enlarged data stack");
      // FIXME: realloc would be faster.
      for (integer_t i = 0; i < data_stack_size; i++)
	stack[i] = data_stack[i];

      mem_free (data_stack);
      data_stack = stack;
    }
#endif

#ifdef PWG_DEBUG
  trace_note (obj, "push");
#endif
  data_stack[data_stack_size++] = obj;
}

pwg_t
pop_data (void)
{
#ifndef PWG_FAST
  if (data_stack_size == 0)
    err ("stack underflow");
#endif
  pwg_t result = data_stack[--data_stack_size];
#ifdef PWG_DEBUG
  trace_note (result, "pop");
#endif
  return result;
}

static void
closure_call (pwg_t sub)
{
  /* I tried hard to make this readable. This code avoids deep
   * nestings. Instead, it handles each case on its own, in the order
   * of how often they occur in practice. So each case here is quite
   * trivial, but we do have a bit of duplication, which we reduce by
   * using the following macros.
   */
#define EVAL_RET() do { eval (code, env); return; } while (0)
#define ADD2ENV(x) env = pwg_cons ((x), env)

#define ADDARGS2ENV(lst) LIST_FOR_EACH ((lst), ptr)			\
    env = pwg_cons (pwg_cons (pwg_car (ptr), pop_data ()), env);

#define ADDRE2ENV(re_data, re_env) do {					\
    pwg_t tmp = pwg_cons (re_sym, PWG_VOID); /* Filled below. */	\
    ADD2ENV (tmp);							\
    pwg_set_cdr (tmp, pwg_closure_new (re_data, re_env));		\
  } while (0)

#define ADDSPC2ENV() do {				\
    pwg_t new_space = pwg_hash_new (sint (4));		\
    pwg_hash_set_default (new_space, PWG_NONE);		\
    ADD2ENV (new_space);				\
  } while (0)

  pwg_t code;
  pwg_t env = pwg_closure_env (sub);
  pwg_t data = pwg_closure_data (sub);

  /* (...) */
  if (!unbooled (pwg_proc_p (data)))
    {
      code = data; /* The most beautiful line of code. */
      EVAL_RET ();
    }

  code = pwg_proc_code (data);
  bool open_space_p = unbooled (pwg_proc_open_space_p (data));
  bool re_p = unbooled (pwg_proc_re_p (data));

  /* |Args(...) */
  if (!re_p && !open_space_p)
    {
      ADDARGS2ENV (pwg_proc_args (data));
      EVAL_RET ();
    }

  pwg_t extra = pwg_proc_extra_args (data);
  bool extra_p = !eq_p (extra, PWG_NIL);

  /* @(...) and @Args(...) */
  if (re_p && !extra_p && !open_space_p)
    {
      ADDARGS2ENV (pwg_proc_args (data));
      ADDRE2ENV (code, env);
      EVAL_RET ();
    }

  /* %(...) and %Args(...) */
  if (open_space_p && !re_p)
    {
      ADDSPC2ENV ();
      ADDARGS2ENV (pwg_proc_args (data));
      EVAL_RET ();
    }

  /* @Args|Extra(...) and @|Extra(...) */
  if (re_p && extra_p && !open_space_p)
    {
      ADDARGS2ENV (pwg_proc_args (data));
      ADDRE2ENV (pwg_proc_new (code, pwg_proc_extra_args (data), PWG_NIL,
			       PWG_FALSE, PWG_FALSE),
		 env);
      ADDARGS2ENV (pwg_proc_extra_args (data));
      EVAL_RET ();
    }
  // FIXME!
  err ("internal error: this kind of closure is not implemented yet");

  /* %@(...) and %@Args(...) */
  if (re_p && open_space_p && !extra_p)
    {
      // order for this call: re+args+newspace+prev_env
      // order for re-proc (which is %(..)):  re+args+prev_env
      ADDARGS2ENV (pwg_proc_args (data));
      ADDRE2ENV (pwg_proc_new (code, PWG_NIL, PWG_NIL, PWG_FALSE, PWG_TRUE),
		 env);
      ADDSPC2ENV ();
      // FIXME: finish
    }

  /* %@|Extra(...) and %@Args|Extra(...)*/
  if (re_p && open_space_p && extra_p)
      // order for this call: re+extra+args+newspace+prev_env
      // order for re-proc (which is %Extra(..)):  re+args+prev_env
    {}

  err ("internal error: calling an impossible closure");
}

static void
call_sub (pwg_t sub)
{
  switch (mem_ref_type (sub))
    {
    case TYPE_CLOSURE:
      closure_call (sub);
      break;

    case TYPE_PSUB:
      psub_call (sub);
      break;

    default:
      err ("calling something not a closure or psub");
    }
}

/* Does it reference a value or is it a call to a sub? */
static bool
value_ref_p (pwg_t sym)
{
#ifndef PWG_FAST
  mem_ref_obj (sym, TYPE_SYM); /* Typecheck.  */
#endif
  mem_set_type (sym, TYPE_STR);
  bool result = isupper (unchr (pwg_str_ref (sym, sint (0))));
  mem_set_type (sym, TYPE_SYM);
  return result;
}



/* This is an order_iter_t.*/
static void
order_usercode (pwg_t obj, void *hook)
{
  push_data (obj);
  pwg_call (pwg_new (hook));
}

/* Process until the next statement and return the next (or PWG_NIL). */
static pwg_t
process_statement (pwg_t lst, order_iter_t fn, void *hook)
{
  if (eq_p (lst, PWG_NIL))
    return PWG_NIL;
  if (eq_p (pwg_car (lst), PWG_SEPARAT))
    return pwg_cdr (lst);
  pwg_t result = process_statement (pwg_cdr (lst), fn, hook);
  fn (pwg_car (lst), hook);
  return result;
}

void
in_order (pwg_t lst, order_iter_t fn, void *hook)
{
  pwg_t next = lst;

  while (!eq_p (next = process_statement (next, fn, hook), PWG_NIL))
    ;
}

pwg_t
pwg_in_order (pwg_t proc, pwg_t lst)
{
  in_order (lst, order_usercode, (void *) pwg_get (proc));
  return PWG_NONE;
}

static int call_depth = 0;

static inline void
inc_depth (void)
{
  call_depth++;
  if (call_depth >= CALL_STACK_MAX)
    err ("call stack overflow");
}

#define PUSH_CALL(x) call_stack = pwg_cons (x, call_stack); inc_depth ();
#define POP_CALL() call_stack = pwg_cdr (call_stack); call_depth--;

/* This is an order_iter_t.*/
static void
element_eval (pwg_t o, void *hook)
{
  pwg_t env = pwg_new (hook);

  if (PWG_IMMP (o))
    {
      if (eq_p (o, PWG_NIL))
	push_data (pwg_closure_new (o, env));
      else
	push_data (o);
    }
  else
    switch (mem_ref_type (o))
      {
      case TYPE_SYM:
	{
	  pwg_t found = lookup (o, env);

	  if (value_ref_p (o))
	    push_data (found);
	  else
	    {
#ifdef PWG_DEBUG
	      trace_note (o, "call");
#endif
	      int exc;
	      PUSH_CALL (o);
	      EXC_TRY
		call_sub (found);
	      EXC_ENSURE (exc)
		POP_CALL ();
	      EXC_FWD (exc);
	    }
	}
	break;

      case TYPE_CONS: case TYPE_PROC:
	push_data (pwg_closure_new (o, env));
	break;

      case TYPE_QUOTOR:
	push_data (pwg_unwrap (o));
	break;

      case TYPE_SETTOR:
	lookup_and_set (pwg_unwrap (o), env, pop_data ());
	break;

      case TYPE_PUSHOR:
	push_data (lookup (pwg_unwrap (o), env));
	break;

      case TYPE_STR: case TYPE_FLOAT:
      case TYPE_HASH: case TYPE_ARY:
      case TYPE_PSUB: case TYPE_CLOSURE:
      case TYPE_STREAM:
      default: /* Or should we raise an error for unknown types? */
	push_data (o);
	break;
      }
}

/* Evaluate LST in ENV.  */
void
eval (pwg_t lst, pwg_t env)
{
  int exc;
  PUSH_CALL (pwg_cons (lst, env));
  EXC_TRY
    in_order (lst, element_eval, (void *) pwg_get (env));
  EXC_ENSURE (exc)
    POP_CALL ();
  EXC_FWD (exc);
}

/* IMPORTANT FIXME: will the stuff created by pwg_stream_in be GC'd? */
static void
process_stream (pwg_t stream, pwg_t module_name)
{
  eval (pwg_stream_in (stream), pwg_hash_item (modules, module_name));
}

static void
display_error (int exc)
{
  // FIXME: we should use stderr here, i think
  switch (exc)
    {
    case EXC_USER: puts ("Exception raised"); break;
    case EXC_ERR: puts ("Error raised"); break;
    }
  pwg_stream_out (pwg_stream_stdout (), pwg_data_stack ());
  puts (" # Data stack");
#ifndef PWG_FAST
  call_stack_out (pwg_stream_stdout (), pwg_exc_call_stack ());
#endif
  pwg_stream_write (pwg_stream_stdout (), chr ('\n'));
  pwg_stream_out (pwg_stream_stdout (), pwg_exc_last ());
  pwg_stream_write (pwg_stream_stdout (), chr ('\n'));
}

/* Will be added to list of modules, so won't be GCed by accident. */
static pwg_t repl_mod;

void
repl ()
{
#define USE "use: | M (Used-Modules: cons M Used-Modules module-enter M)"
  process_stream (pwg_str_src_new (str_new (USE)), repl_sym);
#undef USE

  int line_cnt = 0;
  while (1)
    {
      int exc;

      pwg_stream_write_str (pwg_stream_stdout (), str_new ("#!cnz!# "));
      EXC_TRY
	{
	  line_cnt++; /* Do it here because reading might throw. */
	  pwg_t line = pwg_stream_read_line (pwg_stream_stdin ());

	  if (eof_p (line))
	    {
	      puts ("");
	      return;
	    }

	  if (eq_p (pwg_str_size (line), sint (0)))
	    data_stack_size = 0;
	  else
	    {
	      pwg_t stream = pwg_str_src_new (line);
	      stream_pwg_set_line (stream, line_cnt);
	      process_stream (stream, repl_sym);
	    }

	  pwg_t stack = pwg_data_stack ();
	  if (!eq_p (stack, PWG_NIL))
	    {
	      pwg_stream_out (pwg_stream_stdout (), stack);
	      pwg_stream_write (pwg_stream_stdout (), chr ('\n'));
	    }
	}
      EXC_ENSURE (exc)
	if (exc != 0)
	  display_error (exc);
      EXC_END;
    }
}

/* Find a module, open and process it.  */
static void
run_module (pwg_t name)
{
  int exc;
  pwg_t src;
  pwg_t fn = pwg_str_append (pwg_sym_to_str (name), str_new (".cnz"));

  EXC_TRY
    src = pwg_file_src_new (fn);
  EXC_ENSURE (exc)
    if (exc != 0)
      {
	EXC_TRY
	  {
	    const char *path = getenv ("PWGPATH");

	    if (path == NULL)
	      path = ""; // FIXME: default path
	    src = pwg_file_src_new (STR_APP3 (str_new (path),
					      str_new ("/"),
					      fn));
	  }
	EXC_ENSURE (exc)
	  if (exc != 0)
	    {
	      exc_store (pwg_str_append
			 (str_new ("module not found in PWGPATH: "), fn));
	      exc_raise (EXC_ERR);
	    }
	EXC_END;
      }
  EXC_END;
  process_stream (src, name);
  pwg_file_src_close (src);
}

// FIXME: should this be done w/ a global var?
pwg_t
pwg_common_modules (void)
{
  return pwg_cons (sym ("core"), pwg_cons (sym ("default"), PWG_NIL));
}

/* Create the module NAME (if it does not exist yet) that uses
   USED_MODULES (a list of module names which have already been
   loaded) and find the corresponding source file, then execute the
   code in the new module.  */
pwg_t
pwg_module_load (pwg_t name, pwg_t used_modules)
{
  if (eq_p (pwg_hash_item (modules, name), PWG_VOID))
    {
      pwg_t m = pwg_hash_new (sint (32));

      pwg_hash_set_default (m, PWG_NONE);
      lookup_and_set (used_modules_sym, m, used_modules);
      pwg_hash_set_item (modules, name, m);
      run_module (name);
    }
  return PWG_NONE;
}

pwg_t
pwg_module_reload (pwg_t name)
{
  run_module (name);
  return PWG_NONE;
}

static pwg_t core_mod;

pwg_t
pwg_core_module (void)
{
  return core_mod;
}



pwg_t
pwg_call (pwg_t sub)
{
  call_sub (sub);
  return PWG_NONE;
}

pwg_t
pwg_if (pwg_t cond, pwg_t then_part, pwg_t else_part)
{
  pwg_call (cond);
  if (unbooled (pop_data ()))
    pwg_call (then_part);
  else
    pwg_call (else_part);
  return PWG_NONE;
}

pwg_t
pwg_while (pwg_t cond, pwg_t body)
{
  pwg_call (cond);
  while (unbooled (pop_data ()))
    {
      pwg_call (body);
      pwg_call (cond);
    }
  return PWG_NONE;
}

pwg_t
pwg_eq_p (pwg_t a, pwg_t b)
{
  return booled (eq_p (a, b));
}

pwg_t pwg_nil (void) { return PWG_NIL; }
pwg_t pwg_eof (void) { return PWG_EOF; }

pwg_t
pwg_dup (void)
{
#ifdef PWG_FAST
  return data_stack[data_stack_size - 1];
#else
  pwg_t o = pop_data (); /* Checks if stack is empty.  */
  push_data (o);
  return o;
#endif
}



static struct exc *excs = NULL;

void
exc_register_internal (struct exc *e)
{
  e->next = excs;
  excs = e;
}

void
exc_unregister_internal (void)
{
  excs = excs->next; 
}

static pwg_t safed_call_stack;
static pwg_t exc_val;

/* Store VAL (by creating a copy) to describe the last exception, then
   jump to the next exception handler.  */
void
exc_store (pwg_t val)
{
  int exc;

  EXC_TRY
    {
      exc_val = pwg_copy (val);
#ifndef PWG_FAST
      safed_call_stack = pwg_shallow_copy (call_stack);
#endif
    }
  EXC_ENSURE (exc)
  {} // FIXME: i think we can avoid that
  EXC_FWD (exc);
}

void
exc_raise (int num)
{
  if (excs == NULL)
    {
      display_error (num);
      abort ();
    }
  longjmp (excs->jb, num);
}

pwg_t
pwg_exc_forward (void)
{
  exc_raise (EXC_USER);
  return PWG_NONE;
}

pwg_t
pwg_exc_last (void)
{
  return exc_val;
}

pwg_t
pwg_exc_call_stack (void)
{
  return safed_call_stack;
}

pwg_t
pwg_exc_raise (pwg_t val)
{
  exc_store (val);
  exc_raise (EXC_USER);
  return PWG_NONE;
}

pwg_t
pwg_exc_try (pwg_t try, pwg_t ensure, pwg_t except)
{
  int which;

  EXC_TRY
    pwg_call (try);
  EXC_ENSURE (which)
    {
      pwg_call (ensure);
      switch (which)
	{
	case 0: break;
	case EXC_USER:
	  pwg_call (except);
	  break;
	case EXC_ERR:
	  pwg_exc_forward ();
	default:
	  err ("invalid exception");
	}
    }
  EXC_END;
  return PWG_NONE;
}

/* Create hash that represents an exception.  */
pwg_t
pwg_exc_hash (pwg_t type, pwg_t msg)
{
  pwg_t ht = pwg_hash_new (sint (4));

  pwg_hash_set_item (ht, sym ("type"), type);
  pwg_hash_set_item (ht, sym ("msg"), msg);
  return ht;
}

void
err (const char *s)
{
  exc_store (str_new (s));
  exc_raise (EXC_ERR);
}

void
exc (const char *type, pwg_t msg)
{
  pwg_exc_raise (pwg_exc_hash (sym (type), msg));
}



void
mark_stack (void)
{
  int i;

  for (i = 0; i < data_stack_size; i++)
    gc_mark (data_stack[i]);

  gc_mark (call_stack);
  gc_mark (safed_call_stack);
  gc_mark (exc_val);
}



void
setup_eval (void)
{
  data_stack_size = 0;
#ifndef PWG_FAST
  data_stack_max = DATA_STACK_INIT;
  data_stack = mem_alloc (sizeof (pwg_t) * data_stack_max,
			  "initial data stack");
#endif
  call_stack = PWG_NIL;
  safed_call_stack = PWG_NIL;

  re_sym = sym ("re");
  core_sym = sym ("core");
  repl_sym = sym ("repl");
  used_modules_sym = sym ("Used-Modules");

  core_mod = pwg_hash_new (sint (200));
  pwg_hash_set_default (core_mod, PWG_NONE);
  lookup_and_set (used_modules_sym, core_mod, PWG_NIL);

  repl_mod = pwg_hash_new (sint (32));
  pwg_hash_set_default (repl_mod, PWG_NONE);
  lookup_and_set (used_modules_sym, repl_mod,
		  pwg_cons (sym ("deps"), pwg_common_modules ()));

  modules = pwg_hash_new (sint (8));
  pwg_hash_set_default (modules, PWG_VOID);
  pwg_hash_set_item (modules, core_sym, core_mod);
  pwg_hash_set_item (modules, repl_sym, repl_mod);

  safed_call_stack = PWG_VOID;
  exc_val = PWG_VOID;
}

static pwg_t
make_args (void)
{
  extern int global_argc;
  extern char **global_argv;
  pwg_t result = PWG_NIL;

  /* The first one is reserved for the source file/dir name.  */
  for (int i = global_argc - 1; i >= 1; i--)
    result = pwg_cons (str_new (global_argv[i]), result);
  return result;
}

void
init_eval (void)
{
  lookup_and_set (sym ("Program-Args"), pwg_core_module (), make_args ());

  def_core_psub ("module-load", pwg_module_load, 2);
  def_core_psub ("module-reload", pwg_module_reload, 1);
  def_core_psub ("core-module", pwg_core_module, 0);
  def_core_psub ("common-modules", pwg_common_modules, 0);

  def_core_psub ("in-order", pwg_in_order, 2);

  def_core_psub ("call", pwg_call, 1);
  def_core_psub ("psub?", pwg_psub_p, 1);
  def_core_psub ("if", pwg_if, 3);
  def_core_psub ("while", pwg_while, 2);
  def_core_psub ("eq?", pwg_eq_p, 2);
  def_core_psub ("nil", pwg_nil, 0);
  def_core_psub ("eof", pwg_eof, 0);
  def_core_psub (".", pwg_dup, 0);

  def_core_psub ("exc-last", pwg_exc_last, 0);
  def_core_psub ("exc-call-stack", pwg_exc_call_stack, 0);
  def_core_psub ("raise", pwg_exc_raise, 1);
  def_core_psub ("exc-forward", pwg_exc_forward, 0);
  def_core_psub ("try", pwg_exc_try, 3);
  def_core_psub ("exc-hash", pwg_exc_hash, 2);
}

void
finish_eval (void)
{
#ifndef PWG_FAST
  mem_free (data_stack);
#endif
}
