/* hash.c -- Hash table implementation.
 * Copyright (C) 2014 Wolfgang Jaehrling
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/* Biggest FIXME of all time:
   There is just one thing wrong about the hash table implementation:
   It's not a hash table implementation, since it does not use a hash
   function. :-) We should integrate parts from symkey.c here.
 */

#include <stdlib.h>
#include <stdio.h>

#include "hash.h"
#include "mem.h"
#include "compound.h"
#include "eval.h"

#define HASHTAB_MAXLOAD 200 /* Value between 0 and 255.  */

struct hash
{
  natural_t size;
  natural_t load;

  natural_t *keys;
  natural_t *vals;

  pwg_t default_value; /* Used for pwg-level hashes only.  */
};

/* Return a new hash of size INITSIZE or the default size if INITSIZE
   is zero.  Use EQUAL_P to compare keys, or compare for identity if
   EQUAL_P is false.  */
PWG_GCC_MALLOC hash_t
hash_new (natural_t initsize)
{
  hash_t ht = mem_alloc (sizeof (struct hash), "hash table");

  ht->size = initsize;
  ht->load = 0;
  ht->keys = mem_alloc (initsize * sizeof (natural_t), "hash table keys");
  ht->vals = mem_alloc (initsize * sizeof (natural_t), "hash table values");
  ht->default_value = PWG_INVAL;

  for (natural_t i = 0; i < ht->size; i++)
    ht->keys[i] = ht->vals[i] = 0;

  return ht;
}

void
hash_free (hash_t ht)
{
  mem_free (ht->keys);
  mem_free (ht->vals);
  mem_free (ht);
}

/* Find the entry in HT with KEY and provide the entry number in *POS.
   Skip over deleted entries iff SKIP_DEL_P is true.  Return true iff
   their was an entry with this key already.  */
static bool
find_slot (hash_t ht, natural_t key, bool skip_del_p, natural_t *pos)
{
  natural_t start = key % ht->size;

  *pos = start;
  do
    {
      /* If we want to skip over deleted items and the current one is
	 deleted, then it is obviously not interesting for us and we
	 can skip checking the key.  */
      bool interesting = true;

      if (ht->vals[*pos] == 0)
	return false;
      if (ht->vals[*pos] == (natural_t) -1)
	{
	  if (!skip_del_p)
	    return false;
	  else
	    interesting = false;
	}
      if (interesting && (ht->keys[*pos] == key))
	return true;
      *pos = (*pos + 1) % ht->size;
    }
  while (*pos != start);

  /* Not reached, unless table is full (which should not happen).  */
  abort ();
  return false;
}

static hash_t
remake_table (hash_t ht, bool enlarge_p)
{
  hash_t new_ht = hash_new (ht->size * (enlarge_p ? 2 : 1));

  for (natural_t i = 0; i < ht->size; i++)
    if (ht->vals[i] != 0 && ht->vals[i] != (natural_t) -1)
      hash_set (new_ht, ht->keys[i], ht->vals[i]);
  return new_ht;
}

static void
enlarge_table (hash_t ht)
{
  hash_t new_ht = remake_table (ht, true);

  /* Kill old table and update pointers.  */
  mem_free (ht->keys);
  mem_free (ht->vals);
  ht->size = new_ht->size;
  ht->keys = new_ht->keys;
  ht->vals = new_ht->vals;
  mem_free (new_ht);
  
}

void
hash_set (hash_t ht, natural_t key, natural_t val)
{
  natural_t pos;

  if (val == 0 || val == (natural_t) -1)
    err ("internal error: invalid hash value");

  if (!find_slot (ht, key, false, &pos))
    {
      /* We are adding a new entry.  */
      ht->load++;
      if (((ht->load << 8) / ht->size) > HASHTAB_MAXLOAD)
	{
	  enlarge_table (ht);
	  find_slot (ht, key, false, &pos);
	}
    }
  ht->keys[pos] = key;
  ht->vals[pos] = val;
}

natural_t
hash_val (hash_t ht, natural_t key)
{
  natural_t pos;

  find_slot (ht, key, true, &pos);
  return ht->vals[pos]; /* Will return 0 if not found.  */
}

void
hash_remove (hash_t ht, natural_t key)
{
  natural_t pos;

  if (find_slot (ht, key, true, &pos))
    {
      ht->vals[pos] = (natural_t) -1;
      ht->load--;
    }
  /* It's okay to remove a nonexistant item from the hash.  */
}

void
hash_iterate (hash_t ht, hash_iter_t fn, void *hook)
{
  for (natural_t i = 0; i < ht->size; i++)
    if (ht->vals[i] != 0 && ht->vals[i] != (natural_t) -1)
      fn (hook, ht->keys[i], ht->vals[i]);
}



DEF_TYPE(pwg_hash_new (pwg_t size), get_hash, hash_t, TYPE_HASH,
	 hash_new (unsint (size)));
DEF_CONSTRUCTOR(pwg_hash_copy (pwg_t o), hash_t, TYPE_HASH,
		remake_table (get_hash (o), false)); 

DEF_PRED (pwg_hash_p, o, mem_is_type_p (o, TYPE_HASH));

pwg_t
pwg_hash_default (pwg_t h)
{
  pwg_t v = (get_hash (h))->default_value;

  if (eq_p (v, PWG_INVAL))
    err ("hash has no default value");
  return v;
}

pwg_t
pwg_hash_has_default_p (pwg_t h)
{
  return booled (!eq_p (get_hash (h)->default_value, PWG_INVAL));
}

pwg_t
pwg_hash_set_default (pwg_t h, pwg_t v)
{
  (get_hash (h))->default_value = v;
  return PWG_NONE;
}

pwg_t
pwg_hash_set_item (pwg_t h, pwg_t k, pwg_t v)
{
  hash_set (get_hash (h), pwg_get (k), pwg_get (v));
  return PWG_NONE;
}

pwg_t
pwg_hash_item (pwg_t h, pwg_t k)
{
  pwg_t result;
  natural_t pos;
  hash_t ht = get_hash (h);

  if (find_slot (ht, pwg_get (k), true, &pos))
    return pwg_new (ht->vals[pos]);
  else
    return pwg_hash_default (h);
}

pwg_t
pwg_hash_remove (pwg_t hash, pwg_t key)
{
  hash_remove (get_hash (hash), pwg_get (key));
  return PWG_NONE;
}

pwg_t
pwg_hash_iterate (pwg_t subr, pwg_t hash)
{
  pwg_t lst = PWG_NIL;
  hash_t ht = get_hash (hash);

  for (natural_t i = 0; i < ht->size; i++)
    if (ht->vals[i] != 0 && ht->vals[i] != (natural_t) -1)
      {
	push_data (pwg_new (ht->vals[i]));
	push_data (pwg_new (ht->keys[i]));
	pwg_call (subr);
	lst = pwg_cons (pop_data (), lst);
      }
  return lst;
}

void
hash_iterate_pwg (pwg_t hash, hash_iter_pwg_t fn, void *hook)
{
  hash_t ht = get_hash (hash);

  for (natural_t i = 0; i < ht->size; i++)
    if (ht->vals[i] != 0 && ht->vals[i] != (natural_t) -1)
      fn (hook, pwg_new (ht->keys[i]), pwg_new (ht->vals[i]));
}



static void
hash_gc_mark_element (void *hook, pwg_t key, pwg_t value)
{
  gc_mark (key);
  gc_mark (value);
}

void
hash_gc_mark (pwg_t ht)
{
  if (unbooled (pwg_hash_has_default_p (ht)))
    gc_mark (pwg_hash_default (ht));

  hash_iterate_pwg (ht, &hash_gc_mark_element, NULL);
}



void
setup_hash (void)
{}

void
init_hash (void)
{
  def_core_psub ("hash-new", pwg_hash_new, 1);
  def_core_psub ("hash-copy", pwg_hash_copy, 1);
  def_core_psub ("hash?", pwg_hash_p, 1);
  def_core_psub ("hash-default", pwg_hash_default, 1);
  def_core_psub ("hash-default=>", pwg_hash_set_default, 2);
  def_core_psub ("hash-has-default?", pwg_hash_has_default_p, 1);
  def_core_psub ("hash-item=>", pwg_hash_set_item, 3);
  def_core_psub ("hash-item", pwg_hash_item, 2);
  def_core_psub ("hash-remove", pwg_hash_remove, 2);
  def_core_psub ("hash-iterate", pwg_hash_iterate, 2);
}

void
finish_hash (void)
{
}
