/* mem.c -- Memory management utillity implementation.
 * Copyright (C) 2014 Wolfgang Jaehrling
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdlib.h>

#include "mem.h"
#include "symkey.h"
#include "eval.h"
#include "stream.h"
#include "compound.h"
#include "hash.h"

#ifdef PWG_DEBUG
#include <stdio.h>
#endif

#ifdef PWG_DEBUG
static hash_t dbg_hash;  /* Store for allocation purposes.  */
static bool dbg_alloc_p; /* Are we currently allocating memory? */

static void
display_mem (void *hook, void *x, const char *str)
{
  fprintf (stderr, "%p: %s\n", x, str);
}
#endif

void
mem_info (void)
{
#ifdef PWG_DEBUG
  hash_iterate (dbg_hash, (hash_iter_t) display_mem, NULL);
#endif
}



/* Return N new bytes which will be used for DESC, which is an
   informal description used for debugging, especially searching
   memory leaks.  */
PWG_GCC_MALLOC void *
mem_alloc (size_t n, const char *desc)
{
  void *tmp = malloc (n);

  if (PWG_GCC_EXPECT_FALSE (tmp == NULL))
    {
      fprintf (stderr, "Oops - run out of memory. Maybe call 'gc'?");
      exit (1);
    }
#ifdef PWG_DEBUG
  if (PWG_GCC_EXPECT_TRUE (!dbg_alloc_p))
    {
      /* Creating a hash entry may allocate memory.  */
      dbg_alloc_p = true;
      hash_set (dbg_hash, (natural_t) tmp, (natural_t) desc);
      dbg_alloc_p = false;
    }
#endif
  return tmp;
}

/* Release the memory at address DATA, which has been returned from a
   previous call to `mem_alloc'.  */
void
mem_free (void *data)
{
#ifdef PWG_DEBUG
  if (PWG_GCC_EXPECT_TRUE (!dbg_alloc_p))
    hash_remove (dbg_hash, (natural_t) data);
#endif
  free (data);
}



struct gc_obj
{
  struct gc_obj *next;
  pwg_type_t type;
  struct {} data; /* A pwg_t points to this data. */
};

static struct gc_obj *gc_objs = NULL;

pwg_t
gc_alloc (pwg_type_t t, size_t n)
{
  struct gc_obj *gco;
  n += sizeof (struct gc_obj);

  gco = malloc (n);
  if (PWG_GCC_EXPECT_FALSE (gco == NULL))
    {
      fprintf (stderr, "Oops - run out of memory. Maybe call 'gc'?");
      exit (1);
    }
  gco->type = t;
  /* Update chain. */
  gco->next = gc_objs;
  gc_objs = gco;

  return pwg_new (&gco->data);
}

// FIXME: this is for debugging the GC only.
void out (stream_t dest, pwg_t obj, natural_t remain);
static stream_t
stream_ref (pwg_t o)
{
  return *((stream_t *) mem_ref_obj (o, TYPE_STREAM));
}

/* Return how many objects where sweeped. */
static unsigned int
gc_sweep (void)
{
  int i = 0;
  struct gc_obj **gcop = &gc_objs;

  while (*gcop != NULL)
    {
      if (GC_MARKED_P ((*gcop)->type))
	{
	  GC_UNSETMARK ((*gcop)->type);
	  /*out (stream_ref (pwg_stream_stdout ()),
	       pwg_new (&((*gcop)->data)),
	       1);
	       puts (" <K");*/
	  gcop = &((*gcop)->next);
	}
      else
	{
	  struct gc_obj *tmp = *gcop;
	  *gcop = tmp->next;

	  /*out (stream_ref (pwg_stream_stdout ()),
	       pwg_new (&tmp->data),
	       1);
	       puts (" <S");*/
	  if (PWG_GCC_EXPECT_FALSE (TYPE_HASD_P (tmp->type)))
	    switch (tmp->type)
	      {
	      case TYPE_SYM: break; /* FIXME: Implement this. */
	      case TYPE_HASH: hash_free (*(hash_t *) &tmp->data); break;
	      case TYPE_STREAM: stream_free (*(stream_t *) &tmp->data); break;
	      }
	  free (tmp);
	  i++;
	}
    }
  return i;
}

/* Recursively mark the object. */
void
gc_mark (pwg_t o)
{
  pwg_type_t t;

  if (PWG_IMMP (o))
    return;
  /* We can't use mem_ref_type() here, as it doesn't tell us the GC flag. */
  t = ((pwg_type_t *) pwg_get (o))[-1];
  if (GC_MARKED_P (t))
    return;
  /*out (stream_ref (pwg_stream_stdout ()), o, 1);
    printf (" %d <M\n", (int) pwg_get (o));*/

  mem_setmark (o);
  switch (t)
    {
    case TYPE_CONS: cons_gc_mark (o); break;
    case TYPE_SYM: break;
    case TYPE_STR: break;
    case TYPE_ARY: ary_gc_mark (o); break;
    case TYPE_FLOAT: break;
    case TYPE_HASH: hash_gc_mark (o); break;
    case TYPE_PROC: proc_gc_mark (o); break;
    case TYPE_PSUB: break;
    case TYPE_STREAM: stream_gc_mark (o); break;
    case TYPE_CLOSURE: closure_gc_mark (o); break;
    }
}

/* Process from all entry points. */
static void
gc_mark_all (void)
{
  mark_modules ();
  mark_stack ();
  mark_stdstream ();
  //mark_symtable (); // we need this only so that '\' won't be freed.
}

pwg_t
pwg_gc (void)
{
  gc_mark_all ();
  return sint (gc_sweep ());
}

// FIXME. static? (& always inline) - then we also don't need the check
void
mem_setmark (pwg_t o)
{
#ifndef PWG_FAST
  if (PWG_IMMP (o))
    err ("type error: immediate value has no type number");
#endif
  GC_SETMARK (*(((pwg_type_t *) pwg_get (o)) - 1));
}

// FIXME. static? (& always inline) - then we also don't need the check
void
mem_unsetmark (pwg_t o)
{
#ifndef PWG_FAST
  if (PWG_IMMP (o))
    err ("type error: immediate value has no type number");
#endif
  GC_UNSETMARK (*(((pwg_type_t *) pwg_get (o)) - 1));
}

/* Return the type of O, without GC mark.  */
pwg_type_t
mem_ref_type (pwg_t o)
{
#ifndef PWG_FAST
  if (PWG_IMMP (o))
    err ("type error: immediate value has no type number");
#endif
  return GC_DROPMARK (*(((pwg_type_t *) pwg_get (o)) - 1));
}

/* Set the type. Keep the GC-mark flag!  */
// FIXME: seems keeping the flag is not required. don't call it in gc anyway
// but that might happen one day when we collect symbols.
void
mem_set_type (pwg_t o, pwg_type_t t)
{
  void **tmp = (void **) pwg_get (o);
  //tmp[-1] = (void *) (t | (1 & (unsigned) tmp[-1]));
  tmp[-1] = (void *) t;
}

/* Check whether O is of type T.  O may be any object.  */
bool
mem_is_type_p (pwg_t o, pwg_type_t type)
{
  if (PWG_IMMP (o) || mem_ref_type (o) != type)
    return false;
  return true;
}

/* Return a pointer to the memory of O, which must be of type TYPE.  */
void *
mem_ref_obj (pwg_t o, pwg_type_t type)
{
#ifndef PWG_FAST
  if (!mem_is_type_p (o, type))
    err ("type error: accessing object with wrong type");
#endif
  return (void *) pwg_get (o);
}

/* The first word of an obj is a pointer, which now points to x. */
void
mem_set_obj (pwg_t o, pwg_type_t type, void *x)
{
  *((void **) mem_ref_obj (o, type)) = x;
}



void
setup_mem (void)
{
#ifdef PWG_DEBUG
  dbg_alloc_p = true;
  dbg_hash = hash_new (128);
  dbg_alloc_p = false;
#endif
}

void
init_mem (void)
{
  def_core_psub ("gc", pwg_gc, 0);
}

#ifdef PWG_DEBUG
static void
display_unfree_mem (void *hook, void *x, const char *str)
{
  // FIXME: look at these, though most are irrelevant
  //fprintf (stderr, "LEAK: %s\n", str);
}
#endif

void
finish_mem (void)
{
#ifdef PWG_DEBUG
  hash_iterate (dbg_hash, (hash_iter_t) display_unfree_mem, NULL);
  dbg_alloc_p = true;
  hash_free (dbg_hash);
  dbg_alloc_p = false;
#endif
}
