/* num.c -- Numbers implementation.
 * Copyright (C) 2014 Wolfgang Jaehrling
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <stdio.h>
#include <stdint.h>

#include "num.h"
#include "mem.h"
#include "eval.h"
#include "compound.h"

pwg_t
num_float (double d)
{
  pwg_t result = gc_alloc (TYPE_FLOAT, sizeof (double));
  double *tmp = mem_ref_obj (result, TYPE_FLOAT);
  *tmp = d;
  return result;
}

double
num_unfloat (pwg_t f)
{
  return *((double *) mem_ref_obj (f, TYPE_FLOAT));
}



DEF_PRED (pwg_num_float_p, o, mem_is_type_p (o, TYPE_FLOAT));
DEF_PRED (pwg_num_sint_p, o, sint_p (o));

pwg_t
pwg_num_sint_to_float (pwg_t x)
{
  return num_float (unsint (x));
}

pwg_t
pwg_num_float_to_sint (pwg_t x)
{
  return sint (*((double *) mem_ref_obj (x, TYPE_FLOAT)));
}



#define DEF_SINT_OPERATION(name, op)		\
pwg_t						\
name (pwg_t a, pwg_t b)				\
{						\
  return sint (unsint (a) op unsint (b));	\
}						\
extern void dummy (void)

DEF_SINT_OPERATION(pwg_num_mod, %);

DEF_SINT_OPERATION(pwg_num_bit_and, &);
DEF_SINT_OPERATION(pwg_num_bit_or, |);
DEF_SINT_OPERATION(pwg_num_bit_xor, ^);

pwg_t
pwg_num_bit_not (pwg_t x)
{
  return sint (~unsint (x));
}

/* Operations which are valid on sints, floats and combinations of
   those.  */
#define DEF_NUM_OPERATION(name, op, rsint, rfloat)	\
static PWG_GCC_ALWAYS_INLINE pwg_t				\
float_ ## name (pwg_t a, pwg_t b)			\
{							\
  double val_a = num_unfloat (a);			\
  double val_b = num_unfloat (b);			\
							\
  return rfloat (val_a op val_b);			\
}							\
pwg_t							\
name (pwg_t a, pwg_t b)					\
{							\
  bool a_float_p = false;				\
  bool b_float_p = false;				\
							\
  if (PWG_GCC_EXPECT_FALSE (!sint_p (a)))			\
    a_float_p = true;					\
  if (PWG_GCC_EXPECT_FALSE (!sint_p (b)))			\
    b_float_p = true;					\
							\
  if (PWG_GCC_EXPECT_TRUE (!a_float_p && !b_float_p))	\
    return rsint (unsint (a) op unsint (b));		\
  if (a_float_p && !b_float_p)				\
    return float_ ## name (a, num_float (unsint (b)));	\
  if (!a_float_p && b_float_p)				\
    return float_ ## name (num_float (unsint (a)), b);	\
  return float_ ## name (a, b);				\
}							\
extern void dummy (void)

/* Relationals.  */
DEF_NUM_OPERATION (pwg_num_equal_p, ==, booled, booled);
DEF_NUM_OPERATION (pwg_num_nequal_p, !=, booled, booled);
DEF_NUM_OPERATION (pwg_num_geq_p, >=, booled, booled);
DEF_NUM_OPERATION (pwg_num_leq_p, <=, booled, booled);
DEF_NUM_OPERATION (pwg_num_gt_p, >, booled, booled);
DEF_NUM_OPERATION (pwg_num_lt_p, <, booled, booled);

/* Arithmetic.  */
DEF_NUM_OPERATION (pwg_num_plus, +, sint, num_float);
DEF_NUM_OPERATION (pwg_num_minus, -, sint, num_float);
DEF_NUM_OPERATION (pwg_num_mul, *, sint, num_float);

pwg_t
pwg_num_div (pwg_t a, pwg_t b)
{
  if (sint_p (a))
    a = num_float (unsint (a));
  if (sint_p (b))
    b = num_float (unsint (b));
  double val_a = *((double *) mem_ref_obj (a, TYPE_FLOAT));
  double val_b = *((double *) mem_ref_obj (b, TYPE_FLOAT));
  if (val_b == 0.0)
    exc ("math", str_new ("division by zero"));
  return num_float (val_a / val_b);
}



void
setup_num (void)
{}

void
init_num (void)
{
  def_core_psub ("float?", pwg_num_float_p, 1);
  def_core_psub ("sint?", pwg_num_sint_p, 1);
  def_core_psub ("sint->float", pwg_num_sint_to_float, 1);
  def_core_psub ("float->sint", pwg_num_float_to_sint, 1);

  def_core_psub ("mod", pwg_num_mod, 2);
  def_core_psub ("bit-and", pwg_num_bit_and, 2);
  def_core_psub ("bit-or", pwg_num_bit_or, 2);
  def_core_psub ("bit-xor", pwg_num_bit_xor, 2);
  def_core_psub ("bit-not", pwg_num_bit_not, 1);

  def_core_psub ("is=?", pwg_num_equal_p, 2);
  def_core_psub ("is<>?", pwg_num_nequal_p, 2);
  def_core_psub ("is>=?", pwg_num_geq_p, 2);
  def_core_psub ("is<=?", pwg_num_leq_p, 2);
  def_core_psub ("is>?", pwg_num_gt_p, 2);
  def_core_psub ("is<?", pwg_num_lt_p, 2);

  def_core_psub ("+", pwg_num_plus, 2);
  def_core_psub ("-", pwg_num_minus, 2);
  def_core_psub ("*", pwg_num_mul, 2);
  def_core_psub ("/", pwg_num_div, 2);
}

void
finish_num (void)
{}
