/* pwg.h -- Conseptizer main header.
 * Copyright (C) 2014 Wolfgang Jaehrling
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef PWG_PWG_H
#define PWG_PWG_H

#define PWG_VERSION "2558: Prototype"

#include <stdint.h>
#include <stdbool.h>

/* Optimizations for GCC.  */
#ifdef __GNUC__

/* Always inline certain functions.  */
#if ((__GNUC__ == 3) && (__GNUC_MINOR__ >= 3)) || (__GNUC__ >= 4)
#define PWG_GCC_ALWAYS_INLINE __attribute__((__always_inline__))
#endif

/* A new pointer.  */
#if (__GNUC__ >= 3)
#define PWG_GCC_MALLOC __attribute__((__malloc__))
#endif

/* Expect expressions to be true or false.  */
#if (__GNUC__ >= 3)
#define PWG_GCC_EXPECT_TRUE(e) __builtin_expect((e), -1)
#define PWG_GCC_EXPECT_FALSE(e) __builtin_expect((e), 0)
#endif

/* Unused variables.  */
#if ((__GNUC__ == 2) && (__GNUC_MINOR__ >= 95)) || (__GNUC__ >= 3)
#define PWG_GCC_UNUSED __attribute__((__unused__))
#endif

#endif /* __GNUC__ */

/* Pessimal implementations for old GCC and other compilers.  */
#ifndef PWG_GCC_ALWAYS_INLINE
#define PWG_GCC_ALWAYS_INLINE inline
#endif
#ifndef PWG_GCC_MALLOC
#define PWG_GCC_MALLOC
#endif
#ifndef PWG_GCC_EXPECT_FALSE
#define PWG_GCC_EXPECT_FALSE(e) (e)
#endif
#ifndef PWG_GCC_EXPECT_TRUE
#define PWG_GCC_EXPECT_TRUE(e) (e)
#endif

/* Defined in eval.c.  */
void err (const char *s);

/* Our generic integer types.  */
typedef uintptr_t natural_t;
typedef intptr_t integer_t;

/* Lists on a stack objects and how to access them.  */
#ifdef PWG_DEBUG

typedef struct { natural_t obj; } pwg_t;
#define PWG_CONST_INIT {0}
#define pwg_get(x) ((x).obj)
#define pwg_new(x) (pwg_new_internal ((natural_t) (x)))

static PWG_GCC_ALWAYS_INLINE pwg_t
pwg_new_internal (natural_t x)
{
  pwg_t p;
  p.obj = x;
  return p;
}

#else /* !PWG_DEBUG */

typedef natural_t pwg_t;
#define PWG_CONST_INIT 0
#define pwg_new(x) ((pwg_t) (x))
#define pwg_get(x) (x)

#endif

static PWG_GCC_ALWAYS_INLINE bool
eq_p (pwg_t a, pwg_t b)
{
  return pwg_get (a) == pwg_get (b);
}

#define PWG_BITEST(val, one, zero) \
 (((pwg_get (val) & one) == one) && (((pwg_get (val) ^ zero) & zero) == zero))

/* Every object satisfies one of these three predicates.  */
#define PWG_REFP(o) (PWG_BITEST ((o), 0, 3))    /* ...00 */
#define PWG_UNUSEDP(o) (PWG_BITEST ((o), 2, 1)) /* ...10 */
#define PWG_IMMP(o) (PWG_BITEST ((o), 1, 0))     /* ...1 */

/* Special values.  */
#define PWG_SPECIAL(i) (pwg_new (((i) << 3) + 7))  /* ...111 */
#define PWG_VOID (PWG_SPECIAL (0))    /* Unspecified: $? */
#define PWG_TRUE (PWG_SPECIAL (1))    /* True: $+ */
#define PWG_FALSE (PWG_SPECIAL (2))   /* False: $- */
#define PWG_NIL (PWG_SPECIAL (3))     /* Empty list: () */
#define PWG_EOF (PWG_SPECIAL (4))     /* End of file (unprintable).  */
#define PWG_ANY (PWG_SPECIAL (5))     /* Any value (placeholder): $* */
#define PWG_END (PWG_SPECIAL (6))     /* End of stack group: $$ */
#define PWG_SEPARAT (PWG_SPECIAL (7)) /* Statement separator: ; */

/* Used internally only. */
#define PWG_NONE (PWG_SPECIAL (50))    /* No value, not even void. */
#define PWG_INVAL (PWG_SPECIAL (51))   /* Invalid (unprintable) */

#define PWG_INTERNAL_ENDLIST (PWG_SPECIAL (100))
#define PWG_INTERNAL_IMPROPER (PWG_SPECIAL (101))

/* Used by parser, unprintable.  */



static PWG_GCC_ALWAYS_INLINE bool
eof_p (pwg_t o)
{
  return eq_p (o, PWG_EOF);
}

static PWG_GCC_ALWAYS_INLINE bool
bool_p (pwg_t o)
{
  return eq_p (o, PWG_TRUE) || eq_p (o, PWG_FALSE);
}

/* They had to be named this way due to the conflict with the C99 type
   we are using all over the place.  */
static PWG_GCC_ALWAYS_INLINE pwg_t
booled (bool expr)
{
  return expr ? PWG_TRUE : PWG_FALSE;
}
static PWG_GCC_ALWAYS_INLINE bool
unbooled (pwg_t b)
{
  return !eq_p (b, PWG_FALSE);
}



#define DEF_PRED(name, arg, expr)		\
pwg_t						\
name (pwg_t arg)				\
{						\
  return booled (expr);				\
}						\
extern void dummy (void)

#define DEF_PRED2(name, arg1, arg2, expr)	\
pwg_t						\
name (pwg_t arg1, pwg_t arg2)			\
{						\
  return booled (expr);				\
}						\
extern void dummy (void)



/* Everyone uses sints, so they go here.  */

#define SINT_MAX (INTPTR_MAX >> 2)
/* Always works, and additional work will be optimized away.  */
#define SINT_MIN ((integer_t) (~((~INTPTR_MIN) >> 2)))

static PWG_GCC_ALWAYS_INLINE bool
sint_p (pwg_t o)
{
  return PWG_BITEST (o, 1, 2);  /* ...01 */
}

static PWG_GCC_ALWAYS_INLINE pwg_t
sint (integer_t i)
{
  if (!(i >= SINT_MIN && i <= SINT_MAX))
    err ("number cannot be represented as sint");

  return pwg_new ((i << 2) | 1); /* ...01 */
}

static PWG_GCC_ALWAYS_INLINE integer_t
unsint (pwg_t si)
{
  if (!sint_p (si))
    err ("sint expected");
#if (-1 >> 1) == -1 /* Does bit shifting preserve the sign? */
  return ((intptr_t) pwg_get (si)) >> 2;
#else
  return ((((intptr_t) pwg_get (si)) < 0)
	  ? (~((~((intptr_t) pwg_get (si))) >> 2))
	  : (((intptr_t) pwg_get (si)) >> 2));
#endif
}

#define TOWARD(i, upper)			\
  for (pwg_t i = sint (0);			\
       unsint (i) < unsint (upper);		\
       i = sint (unsint (i) + 1))

#define TOWARD_ZERO(i, upper)			\
  for (pwg_t i = sint (unsint (upper) - 1);	\
       unsint (i) >= 0;				\
       i = sint (unsint (i) - 1))



/* Layout of an objects type field:
   bit 0: mark-flag of GC for the obj
   bit 1: types has_destructor flag 
   remaining: type id
 */

#define TYPE(n) ((n) * 4)
/* Type with destructor: */
#define TYPE_WD(n) (TYPE (n) + 2)
#define TYPE_HASD_P(n) ((n) & 2)
/* GC marking: */
#define GC_MARKED_P(n) ((n) & 1)
#define GC_SETMARK(n) ((n) |= 1)
#define GC_DROPMARK(n) ((n) & ~1)
#define GC_UNSETMARK(n) ((n) &= ~1)

/* Non-immediate types.  */
typedef natural_t pwg_type_t;

#define TYPE_CONS (TYPE (1))
#define TYPE_SYM (TYPE_WD (2))
#define TYPE_STR (TYPE (3))
#define TYPE_ARY (TYPE (4))
#define TYPE_FLOAT (TYPE (5))
#define TYPE_HASH (TYPE_WD (6))
#define TYPE_PROC (TYPE (7))
#define TYPE_PSUB (TYPE (8))
#define TYPE_STREAM (TYPE_WD (9))
#define TYPE_CLOSURE (TYPE (10))

/* Start value for wrapped objects. */
#define END_BASE_TYPES 11

#define TYPE_QUOTOR (TYPE (11))    /* 'foo */
#define TYPE_SETTOR (TYPE (12))    /* foo: */
#define TYPE_PUSHOR (TYPE (13))    /* ^foo */

/* Start value for types added at runtime.  */
#define END_BUILTIN_TYPES 14

#endif /* PWG_PWG_H */
