/* stream.c -- Input/Output stream implementation.
 * Copyright (C) 2014 Wolfgang Jaehrling
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#define _ISOC9X_SOURCE 1 /* snprintf() */
#include <stdio.h>
#include <stdlib.h>

#include <ctype.h>
#include <string.h>
#include <errno.h>

#include "stream.h"
#include "mem.h"
#include "compound.h"
#include "hash.h"
#include "num.h"
#include "symkey.h"
#include "eval.h"

/* Characters valid in directly printable syms in addition to
   alphanumerics.  */
#define SYM_EXTRA_CHRS "-.?+>*![]/<=_"

struct stream_char
{
  pwg_t c;
  struct stream_char *next;
};

struct stream
{
  void *hook;
  pwg_t name;
  integer_t input_line;
  stream_type_t type;
  stream_readfn_t readfn;
  stream_writefn_t writefn;
  stream_markfn_t markfn;
  stream_destructfn_t destructfn;
  struct stream_char *pending;
};

static void
exc_input (stream_t s, const char *desc)
{
  pwg_t msg = pwg_str_dst_new ();

  if (!eq_p (s->name, PWG_VOID))
    pwg_stream_write_str (msg, s->name);
  else
    pwg_stream_write_str (msg, str_new ("<stream>"));
  pwg_stream_write (msg, chr (':'));
  pwg_stream_out (msg, sint (s->input_line));
  pwg_stream_write_str (msg, str_new (": "));
  pwg_stream_write_str (msg, str_new (desc));

  exc ("io", pwg_str_dst_to_str (msg));
}

static void
exc_file_open (pwg_t name)
{
  exc ("io", STR_APP3 (name, str_new (" could not be opened: "),
		       str_new (strerror (errno))));
}



static stream_t
make_stream (void *hook, pwg_t name, stream_type_t type,
	     stream_readfn_t readfn, stream_writefn_t writefn,
	     stream_markfn_t markfn, stream_destructfn_t destructfn)
{
  stream_t s = mem_alloc (sizeof (struct stream), "stream");

  s->hook = hook;
  s->name = name;
  s->input_line = 1;
  s->type = type;
  s->readfn = readfn;
  s->writefn = writefn;
  s->markfn = markfn;
  s->destructfn = destructfn;
  s->pending = NULL;
  return s;
}

void
stream_free (stream_t s)
{
  if (s->destructfn != NULL && s->hook != NULL)
    s->destructfn (s->hook);
  while (s->pending != NULL)
    {
      void *sc = s->pending;
      s->pending = s->pending->next;
      mem_free (sc);
    }
  mem_free (s);
}

pwg_t
stream_read (stream_t s)
{
  pwg_t c;

  if (s->pending != NULL)
    {
      struct stream_char *sc = s->pending;
      c = sc->c;
      s->pending = sc->next;
      mem_free (sc);
    }
  else if (s->readfn != NULL)
    c = s->readfn (s->hook);
  else
    err ("stream supports no input");

  if (eq_p (c, chr ('\n')))
    s->input_line++;

  return c;
}

void
stream_ungetc (stream_t s, pwg_t c)
{
  struct stream_char *sc = mem_alloc (sizeof (struct stream_char), "ungetc");

  if (!eof_p (c) && unchr (c) == '\n') /* Don't call unchr() if it's eof. */
    s->input_line--;

  sc->c = c;
  sc->next = s->pending;
  s->pending = sc;
}

pwg_t
stream_lookahead (stream_t s)
{
  pwg_t c = stream_read (s);
  stream_ungetc (s, c);
  return c;
}

void
stream_write (stream_t s, pwg_t c)
{
  if (s->writefn != NULL)
    s->writefn (s->hook, c);
  else
    stream_ungetc (s, c);
}

void
stream_write_str (stream_t s, const char *p)
{
  for (; *p != '\0'; p++)
    stream_write (s, chr (*p));
}

DEF_TYPE(stream_new (void *hook, pwg_t name, stream_type_t type,
		     stream_readfn_t readfn, stream_writefn_t writefn,
		     stream_markfn_t markfn, stream_destructfn_t destructfn),
	 stream_ref, struct stream *, TYPE_STREAM,
	 make_stream (hook, name, type, readfn, writefn, markfn, destructfn));

void
stream_pwg_set_line (pwg_t obj, int n)
{
  stream_t s = stream_ref (obj);
  s->input_line = n;
}

pwg_t
pwg_stream_read (pwg_t obj)
{
  return stream_read (stream_ref (obj));
}

pwg_t
pwg_stream_lookahead (pwg_t obj)
{
  return stream_lookahead (stream_ref (obj));
}

pwg_t
pwg_stream_write (pwg_t obj, pwg_t c)
{
  stream_write (stream_ref (obj), c);
  return PWG_NONE;
}

pwg_t
pwg_stream_write_str (pwg_t stream, pwg_t str)
{
  char *tmp = unstr (str);
  stream_write_str (stream_ref (stream), tmp);
  mem_free (tmp);
  return PWG_NONE;
}

pwg_t
pwg_stream_ungetc (pwg_t obj, pwg_t c)
{
  stream_ungetc (stream_ref (obj), c);
  return PWG_NONE;
}

pwg_t
pwg_stream_name (pwg_t obj)
{
  return (stream_ref (obj))->name;
}

pwg_t
pwg_stream_input_line (pwg_t obj)
{
  return sint ((stream_ref (obj))->input_line);
}

pwg_t
pwg_stream_read_line (pwg_t stream)
{
  pwg_t lst = PWG_NIL;

  while (1)
    {
      pwg_t c = pwg_stream_read (stream);

      if (eof_p (c))
	{
	  if (eq_p (lst, PWG_NIL))
	    return PWG_EOF;
	  else
	    c = chr ('\n');
	}
      if (eq_p (c, chr ('\n')))
	return pwg_list_to_str (pwg_reverse_x (lst));
      lst = pwg_cons (c, lst);
    }
  return PWG_VOID; /* Not reached.  */
}

void
stream_gc_mark (pwg_t o)
{
  stream_t s = stream_ref (o);
  gc_mark (s->name);
  if (s-> markfn != NULL)
    s->markfn (s);
}



struct str_src
{
  pwg_t str;
  integer_t pos;
};

static void
str_src_markfn (struct str_src *sis)
{
  gc_mark (sis->str);
}

static pwg_t
str_src_readfn (struct str_src *sis)
{
  if (sis->pos == unsint (pwg_str_size (sis->str)))
    return PWG_EOF;
  else
    return pwg_str_ref (sis->str, sint (sis->pos++));
}

pwg_t
pwg_str_src_new (pwg_t str)
{
  struct str_src *sis = mem_alloc (sizeof (struct str_src), "str src");
  pwg_t obj = stream_new (sis, PWG_VOID, STREAM_STR_IN,
			  (stream_readfn_t) str_src_readfn, NULL,
			  (stream_markfn_t) str_src_markfn,
			  (stream_destructfn_t) mem_free);

  sis->str = str;
  sis->pos = 0;
  return obj;
}





struct str_dst
{
  pwg_t lst;
};

static void
str_dst_markfn (struct str_dst *sos)
{
  gc_mark (sos->lst);
}

static void
str_dst_writefn (struct str_dst *sos, pwg_t c)
{
  sos->lst = pwg_cons (c, sos->lst);
}

pwg_t
pwg_str_dst_new (void)
{
  struct str_dst *sos = mem_alloc (sizeof (struct str_dst), "str dst");
  pwg_t obj = stream_new (sos, PWG_VOID, STREAM_STR_OUT, NULL,
			  (stream_writefn_t) str_dst_writefn,
			  (stream_markfn_t) str_dst_markfn,
			  (stream_destructfn_t) mem_free);

  sos->lst = PWG_NIL;
  return obj;
}

pwg_t
pwg_str_dst_to_str (pwg_t s)
{
  stream_t stream = stream_ref (s);
  struct str_dst *sos = stream->hook;

  if (stream->type != STREAM_STR_OUT)
    err ("type error: not a str-dst");
  return pwg_list_to_str (pwg_reverse (sos->lst));
}



struct file_src
{
  FILE *fp;
};

static void
file_src_free (struct file_src *fis)
{
  if (fis->fp != NULL)
    {
      fclose (fis->fp);
      fis->fp = NULL; //FIXME: makes no sense, we free fis anyway
    }
  mem_free (fis);
}

static pwg_t
file_src_readfn (struct file_src *fis)
{
  int i = getc (fis->fp);
  return (i == EOF) ? PWG_EOF : chr (i);
}

pwg_t
pwg_file_src_close (pwg_t obj)
{
  stream_t s = stream_ref (obj);
  // FIXME: should call destructor function and be generic.
  file_src_free ((struct file_src *) (s->hook));
  s->hook = NULL;
  return PWG_NONE;
}

pwg_t
pwg_file_src_new (pwg_t filename)
{
  char *fn = unstr (filename);
  struct file_src *fis = mem_alloc (sizeof (struct file_src), "file src");
  FILE *fp = fopen (fn, "r");

  mem_free (fn);
  if (fp == NULL)
    exc_file_open (filename);
  fis->fp = fp;

  return stream_new (fis, pwg_str_copy (filename), STREAM_FILE_IN,
		     (stream_readfn_t) file_src_readfn, NULL, NULL,
		     (stream_destructfn_t) file_src_free);
}



struct file_dst
{
  FILE *fp;
};

static void
file_dst_free (struct file_dst *fos)
{
  if (fos->fp != NULL)
    {
      fclose (fos->fp);
      fos->fp = NULL;
    }
  mem_free (fos);
}

static void
file_dst_writefn (struct file_dst *fos, pwg_t c)
{
  putc (unchr (c), fos->fp);
}

pwg_t
pwg_file_dst_close (pwg_t obj)
{
  stream_t s = stream_ref (obj);
  file_dst_free ((struct file_dst *) (s->hook));
  s->hook = NULL;
  return PWG_NONE;
}

pwg_t
pwg_file_dst_new (pwg_t filename, pwg_t append_p)
{
  char *fn = unstr (filename);
  struct file_dst *fos = mem_alloc (sizeof (struct file_dst), "file dst");
  FILE *fp = fopen (fn, unbooled (append_p) ? "w+" : "w");

  mem_free (fn);
  if (fp == NULL)
    exc_file_open (filename);
  fos->fp = fp;

  return stream_new (fos, pwg_str_copy (filename), STREAM_FILE_OUT, NULL,
		     (stream_writefn_t) file_dst_writefn, NULL,
		     (stream_destructfn_t) file_dst_free);
}



static pwg_t the_stdout;
static pwg_t the_stderr;
static pwg_t the_stdin;

void
mark_stdstream (void)
{
  gc_mark (the_stdout);
  gc_mark (the_stderr);
  gc_mark (the_stdin);
}

static struct file_dst stdout_hook;
static struct file_dst stderr_hook;
static struct file_src stdin_hook;

static struct stream stdout_stream =
  { &stdout_hook, PWG_CONST_INIT, 1, STREAM_FILE_OUT, NULL,
    (stream_writefn_t) file_dst_writefn, NULL, NULL, NULL };
static struct stream stderr_stream =
  { &stderr_hook, PWG_CONST_INIT, 1, STREAM_FILE_OUT, NULL,
    (stream_writefn_t) file_dst_writefn, NULL, NULL, NULL };
static struct stream stdin_stream =
  { &stdin_hook, PWG_CONST_INIT, 1, STREAM_FILE_IN,
    (stream_readfn_t) file_src_readfn, NULL, NULL, NULL, NULL };

stream_t
stream_stdout (void)
{
  return &stdout_stream;
}

stream_t
stream_stderr (void)
{
  return &stderr_stream;
}

stream_t
stream_stdin (void)
{
  return &stdin_stream;
}

static pwg_t
pwgize_stream (stream_t s)
{
  pwg_t result = gc_alloc (TYPE_STREAM, sizeof (void *));
  *((stream_t *) mem_ref_obj (result, TYPE_STREAM)) = s;
  return result;
}

pwg_t
pwg_stream_stdout (void)
{
  return the_stdout;
}

pwg_t
pwg_stream_stderr (void)
{
  return the_stderr;
}

pwg_t
pwg_stream_stdin (void)
{
  return the_stdin;
}



/* Rule for output functions: When abbreviating, ensure that you
 * invoke the `register_object_out' function where it is needed.
 */

//static FIXME
void out (stream_t dest, pwg_t obj, natural_t remain);

static void
out_int_rec (stream_t dest, integer_t i)
{
  if (i == 0)
    return;
  out_int_rec (dest, i / 10);
  stream_write (dest, chr ('0' + (i % 10)));
}

static void
out_int (stream_t dest, integer_t i)
{
  if (i == 0)
    {
      stream_write (dest, chr ('0'));
      return;
    }
  if (i < 0)
    {
      stream_write (dest, chr ('-'));
      i = -i;
    }
  out_int_rec (dest, i);
}

static void
out_sint (stream_t dest, pwg_t obj)
{
  out_int (dest, unsint (obj));
}

/* Control characters.  */
static const char *chr_names[256] = {
  [0] = "nul", [1] = "soh", [2] = "stx", [3] = "etx", [4] = "eot", [5] = "enq",
  [6] = "ack", [7] = "alert", [8] = "backspace", [9] = "tab", [10] = "newline",
  [11] = "vtab", [12] = "form-feed", [13] = "line-feed", [14] = "so",
  [15] = "si", [16] = "dle", [17] = "dc1", [18] = "dc2", [19] = "dc3",
  [20] = "dc4", [21] = "nak", [22] = "syn", [23] = "etb", [24] = "can",
  [25] = "em", [26] = "sub", [27] = "esc", [28] = "fs", [29] = "gs",
  [30] = "rs", [31] = "us", /* [92] = "backslash", */ [255] = "del"
};

static void
out_chr (stream_t dest, pwg_t obj)
{
  natural_t c = unchr (obj);

  stream_write (dest, chr ('$'));
  stream_write (dest, chr ('\''));
  if (PWG_GCC_EXPECT_FALSE (c == '\''))
    stream_write_str (dest, "quote");
  else if (PWG_GCC_EXPECT_FALSE (c < 256 && chr_names[c] != NULL))
    stream_write_str (dest, chr_names[c]);
  else
    stream_write (dest, obj);
  stream_write (dest, chr ('\''));
}

static void
out_cons (stream_t dest, pwg_t obj, natural_t remain)
{
  bool first = true;

  for (; unbooled (pwg_cons_p (obj)); obj = pwg_cdr (obj))
    {
      if (first)
	{
	  stream_write (dest, chr ('('));
	  first = false;
	}
      else if (!eq_p (pwg_car (obj), PWG_SEPARAT))
	stream_write (dest, chr (' '));
      out (dest, pwg_car (obj), remain - 1);
    }
  if (!eq_p (obj, PWG_NIL))
    {
      stream_write (dest, chr (' '));
      stream_write (dest, chr ('\\'));
      stream_write (dest, chr (' '));
      out (dest, obj, remain - 1);
    }
  stream_write (dest, chr (')'));
}

static void
out_float (stream_t dest, pwg_t obj)
{
  char ptr[32];

  snprintf (ptr, 30, "%g", num_unfloat (obj));
  if (strchr (ptr, '.') == NULL)
    strcat (ptr, ".0");
  stream_write_str (dest, ptr);
}

static void
out_ary (stream_t dest, pwg_t obj, natural_t remain)
{
  pwg_t len = pwg_ary_size (obj);

  stream_write_str (dest, "$(");
  TOWARD (i, len)
    {
      out (dest, pwg_ary_item (obj, i), remain - 1);
      if (!eq_p (i, sint (unsint (len) - 1)))
	stream_write (dest, chr (' '));
    }
  stream_write (dest, chr (')'));
}

static void
out_str (stream_t dest, pwg_t obj)
{
  stream_write (dest, chr ('"'));
  TOWARD (i, pwg_str_size (obj))
    {
      pwg_t c = pwg_str_ref (obj, i);

      if (PWG_GCC_EXPECT_FALSE (unchr (c) == '"'))
	stream_write_str (dest, "$'\"'");
      else if (PWG_GCC_EXPECT_FALSE (unchr (c) == '$'))
	stream_write_str (dest, "$'$'");
      else if (PWG_GCC_EXPECT_FALSE (unchr (c) < 256
				 && chr_names[unchr (c)] != NULL))
	out_chr (dest, c);
      else
	stream_write (dest, c);
    }
  stream_write (dest, chr ('"'));
}

static bool
printable_sym_p (pwg_t str)
{
  bool start_p = true;
  TOWARD (i, pwg_str_size (str))
    {
      integer_t c = unchr (pwg_str_ref (str, i));
#define is_ascii(c) (((c) & ~0x7f) == 0) /* Should work with Unicode.  */
      if (!((is_ascii (c) && (start_p ? isalpha (c) : isalnum (c)))
#undef is_ascii
	    || strchr (SYM_EXTRA_CHRS, c) != NULL))
	return false;
      start_p = false;
    }
  return true;
}

static void
out_sym (stream_t dest, pwg_t obj)
{
  mem_set_type (obj, TYPE_STR);
  if (printable_sym_p (obj))
    TOWARD (i, pwg_str_size (obj))
      stream_write (dest, pwg_str_ref (obj, i));
  else
    {
      stream_write (dest, chr ('$'));
      out_str (dest, obj);
    }
  mem_set_type (obj, TYPE_SYM);
}

struct out_hash
{
  natural_t remain;
  stream_t dest;
};

static void
out_hash_entry (struct out_hash *oh, pwg_t key, pwg_t val)
{
  stream_write_str (oh->dest, " (");
  out (oh->dest, key, oh->remain);
  stream_write_str (oh->dest, " \\ ");
  out (oh->dest, val, oh->remain);
  stream_write (oh->dest, chr (')'));
}

static void
out_hash (stream_t dest, pwg_t obj, natural_t remain)
{
  struct out_hash oh = { remain - 1, dest };

  stream_write_str (dest, "$<");
  if (unbooled (pwg_hash_has_default_p (obj)))
    {
      stream_write_str (dest, "$+ ");
      out (dest, pwg_hash_default (obj), remain - 1);
    }
  else
    stream_write_str (dest, "$-");
  hash_iterate_pwg (obj, (hash_iter_pwg_t) out_hash_entry, &oh);
  stream_write (dest, chr ('>'));
}

static void
out_psub (stream_t dest, pwg_t obj)
{
  stream_write_str (dest, "${psub ");
  out_int (dest, (integer_t) psub_fn (obj));
  stream_write (dest, chr (' '));
  out_int (dest, (integer_t) psub_args (obj));
  stream_write (dest, chr ('}'));
}

static void
out_proc (stream_t dest, pwg_t obj, natural_t remain)
{
  bool opened = false;

  if (unbooled (pwg_proc_open_space_p (obj)))
    {
      stream_write (dest, chr ('%'));
      opened = true;
    }
  if (unbooled (pwg_proc_re_p (obj)))
    {
      stream_write (dest, chr ('@'));
      opened = true;
    }

  if (!eq_p (pwg_proc_args (obj), PWG_NIL))
    {
      if (!opened)
	stream_write (dest, chr ('|'));
      stream_write (dest, chr (' '));
      LIST_FOR_EACH (pwg_proc_args (obj), ptr)
	{
	  out_sym (dest, pwg_car (ptr));
	  stream_write (dest, chr (' '));
	}
    }
  if (!eq_p (pwg_proc_extra_args (obj), PWG_NIL))
    {
      stream_write_str (dest, "| ");
      LIST_FOR_EACH (pwg_proc_extra_args (obj), ptr)
	{
	  out_sym (dest, pwg_car (ptr));
	  stream_write (dest, chr (' '));
	}
    }

  out (dest, pwg_proc_code (obj), remain - 1);
}

static void
out_stream (stream_t dest, pwg_t obj, natural_t remain)
{
  stream_write_str (dest, "${stream ");
  out (dest, pwg_stream_name (obj), remain);
  stream_write (dest, chr (' '));
  out_int (dest, (integer_t) *((void **)
			       mem_ref_obj (obj, TYPE_STREAM)));
  stream_write (dest, chr ('}'));
}

static void
out_closure (stream_t dest, pwg_t obj, natural_t remain)
{
  pwg_t x = pwg_closure_data (obj);
  pwg_t env = pwg_closure_env (obj);

  stream_write_str (dest, "${closure ");
  out (dest, x, remain - 1);
  stream_write (dest, chr (' '));
  out_int (dest, pwg_get (env));
  stream_write (dest, chr ('}'));
}

static void
out_quotor (stream_t dest, pwg_t obj, natural_t remain)
{
  stream_write (dest, chr ('\''));
  out (dest, pwg_unwrap (obj), remain - 1);
}

static void
out_settor (stream_t dest, pwg_t obj, natural_t remain)
{
  out (dest, pwg_unwrap (obj), remain - 1);
  stream_write (dest, chr (':'));
}

static void
out_pushor (stream_t dest, pwg_t obj, natural_t remain)
{
  stream_write (dest, chr ('^'));
  out (dest, pwg_unwrap (obj), remain - 1);
}

static void
out_accessor (stream_t dest, pwg_t obj, natural_t remain)
{
  stream_write (dest, chr (','));
  out (dest, pwg_unwrap (obj), remain - 1);
}

static void
out_accsettor (stream_t dest, pwg_t obj, natural_t remain)
{
  stream_write (dest, chr (','));
  out (dest, pwg_unwrap (obj), remain - 1);
  stream_write (dest, chr (':'));
}

typedef void (*out_fn_t) (stream_t, pwg_t, natural_t);

static void
out_rec (out_fn_t fn, stream_t dest, pwg_t obj, natural_t remain)
{
  if (remain != 0)
    fn (dest, obj, remain);
  else
    stream_write_str (dest, "${...}");
}

/* Print OBJ to DST, with a nesting level of REMAIN. */
// FIXME: should be called DEPTH, not REMAIN.
//static void
void
out (stream_t dest, pwg_t obj, natural_t remain)
{
  if (PWG_IMMP (obj))
    {
      if (sint_p (obj))
	out_sint (dest, obj);
      else if (chr_p (obj))
	out_chr (dest, obj);
      else if (eq_p (obj, PWG_SEPARAT)) stream_write_str (dest, ";");
      else if (eq_p (obj, PWG_NIL)) stream_write_str (dest, "()");
      else if (eq_p (obj, PWG_TRUE)) stream_write_str (dest, "$+");
      else if (eq_p (obj, PWG_FALSE)) stream_write_str (dest, "$-");
      else if (eq_p (obj, PWG_VOID)) stream_write_str (dest, "$?");
      else if (eq_p (obj, PWG_ANY)) stream_write_str (dest, "$*");
      else if (eq_p (obj, PWG_EOF)) stream_write_str (dest, "${eof}");
      else if (eq_p (obj, PWG_INVAL)) stream_write_str (dest, "${inval}");
      else if (eq_p (obj, PWG_NONE)) stream_write_str (dest, "${-}");
      else if (eq_p (obj, PWG_END)) stream_write_str (dest, "$$");
      else err ("internal error: invalid immediate value");
    }
  else
    {
      switch (mem_ref_type (obj))
	{
	case TYPE_CONS: out_rec (out_cons, dest, obj, remain); break;
	case TYPE_SYM: out_sym (dest, obj); break;
	case TYPE_STR: out_str (dest, obj); break;
	case TYPE_ARY: out_rec (out_ary, dest, obj, remain); break;
	case TYPE_FLOAT: out_float (dest, obj); break;
	case TYPE_HASH: out_rec (out_hash, dest, obj, remain); break;
	case TYPE_PROC: out_rec (out_proc, dest, obj, remain); break;
	case TYPE_PSUB: out_psub (dest, obj); break;
	case TYPE_STREAM: out_stream (dest, obj, remain); break;
	case TYPE_CLOSURE: out_rec (out_closure, dest, obj, remain); break;
	case TYPE_QUOTOR: out_rec (out_quotor, dest, obj, remain); break;
	  /* For the following two, we use out_rec in case someone
	     constructs an (invalid) object with 'wrap' that does not
	     contain a sym. */
	case TYPE_SETTOR: out_rec (out_settor, dest, obj, remain); break;
	case TYPE_PUSHOR: out_rec (out_pushor, dest, obj, remain); break;
	  // FIXME: print user-defined types.
	default: err ("internal error: invalid type on output");
	}
    }
}

pwg_t
pwg_stream_out (pwg_t stream, pwg_t obj)
{
  // FIXME: there should be a way to set this (5).
  out (stream_ref (stream), obj, 5);
  return PWG_NONE;
}



/* The Reader.  */

struct reader_object
{
  pwg_t obj;
  enum { OBJ,         /* Just an obj.  */
	 SPECIAL,     /* Parens & Co. */
	 XEOF         /* End of file.  */
  } type;
};

/* Read chrs from src until reaching end of file or ISVALID returns
   zero when applied to the chr. */
static pwg_t
read_into_lst (stream_t src, int (*isvalid) (int))
{
  pwg_t c, lst = PWG_NIL;

  for (c = stream_read (src);
       !eof_p (c) && isvalid (unchr (c));
       c = stream_read (src))
    lst = pwg_cons (c, lst);

  if (!eof_p (c))
    stream_ungetc (src, c);
  return pwg_reverse_x (lst);
}

static pwg_t
read_into_str (stream_t src, int (*isvalid) (int))
{
  return pwg_list_to_str (read_into_lst (src, isvalid));
}

static int
is_not_quote (int c)
{
  return c != '\'';
}

static int
is_symchr_or_asgn (int c)
{
  return isalnum (c) || strchr (SYM_EXTRA_CHRS ":", c) != NULL;  
}

/* Lexer.  */
static struct reader_object
read_next_ro (stream_t src)
{
#define LONG_COMMENT "$:!<" /* No "(", as that easiely happens by... */
#define END_COMMENT  "$:!>" /* ...accident when commenting out a line. */
#define RETURN_RO(o, t)				\
  do {						\
    struct reader_object result;		\
    result.obj = o;				\
    result.type = t;				\
    return result;				\
  } while (0)

  pwg_t c;

  /* Get first non-whitespace char, ignoring comments.  */
  do {
    c = stream_read (src);
    if (eof_p (c))
      RETURN_RO (PWG_EOF, XEOF);
    if (unchr (c) == '#')
      {
	c = stream_read (src);
	if (eof_p (c))
	  RETURN_RO (PWG_EOF, XEOF);

	char *p = strchr (LONG_COMMENT, unchr (c));
	if (p == NULL)
	  /* Read until newline.  */
	  while (unchr (c) != '\n')
	    {
	      c = stream_read (src);
	      if (eof_p (c))
		RETURN_RO (PWG_EOF, XEOF);
	    }
	else
	  /* Read until proper end */
	  while (1)
	    {
	      c = stream_read (src);
	      if (eof_p (c))
		exc_input (src, "eof in #<...># comment");
	      if (unchr (c) == END_COMMENT[p - LONG_COMMENT])
		{
		  c = stream_lookahead (src);
		  if (eof_p (c))
		    exc_input (src, "eof in #<...># comment");
		  if (unchr (c) == '#')
		    {
		      stream_read (src); /* Drop the '#' */
		      c = stream_read (src);
		      if (eof_p (c))
			RETURN_RO (PWG_EOF, XEOF);
		      break;
		    }
		}
	    }
      }
  } while (unchr (c) == ' ' || unchr (c) == '\t'
	   || unchr (c) == '\n' || unchr (c) == 12);

  /* Now we are at the beginning of a token. */
  switch (unchr (c))
    {
    case '"':
      /* String.  */
      {
	pwg_t lst = PWG_NIL;

	for (c = stream_read (src);
	     eof_p (c) || unchr (c) != '"'; /* EOF handled inside loop.  */
	     c = stream_read (src))
	  {
	    if (eof_p (c))
	      exc_input (src, "eof in str");
	    if (unchr (c) == '$')
	      {
		stream_ungetc (src, c);
		struct reader_object r = read_next_ro (src);
		if (r.type != OBJ || !chr_p (r.obj))
		  exc_input (src, "parse error: invalid chr object in str");
		c = r.obj;
	      }
	    lst = pwg_cons (c, lst);
	  }

	RETURN_RO (pwg_list_to_str (pwg_reverse_x (lst)), OBJ);
      }

    case '\'':
      RETURN_RO (chr ('\''), SPECIAL);

    case '^':
      {
	struct reader_object tmp = read_next_ro (src);
	if (tmp.type != OBJ || !unbooled (pwg_sym_p (tmp.obj)))
	  exc_input (src, "a ^ must be followed by a sym");
	RETURN_RO (pwg_pushor_new (tmp.obj), OBJ);
      }

    case ';': RETURN_RO (PWG_SEPARAT, OBJ);

    case '$':
      c = stream_read (src);
      if (eof_p (c))
	exc_input (src, "eof after $");
      switch (unchr (c))
	{
	case '+': RETURN_RO (PWG_TRUE, OBJ);
	case '-': RETURN_RO (PWG_FALSE, OBJ);
	case '?': RETURN_RO (PWG_VOID, OBJ);
	case '*': RETURN_RO (PWG_ANY, OBJ);
	case '$': RETURN_RO (PWG_END, OBJ);

	case '\'':
	  /* Chrs.  */
	  {
	    pwg_t s = read_into_str (src, is_not_quote);

	    c = stream_read (src);
	    if (eof_p (c))
	      exc_input (src, "eof in chr");
	    if (unchr (c) != '\'')
	      err ("internal error: expected '");

	    if (unsint (pwg_str_size (s)) == 1)
	      /* Given directly, extract from str.  */
	      RETURN_RO (pwg_str_ref (s, sint (0)), OBJ);
	    else
	      /* Lookup in table.  */
	      for (int i = 0; i <= 255; i++)
		if (chr_names[i] != NULL && str_eql (s, chr_names[i]))
		  RETURN_RO (chr (i), OBJ);
	    /* Special case '.  */
	    if (str_eql (s, "quote"))
	      RETURN_RO (chr ('\''), OBJ);
	    exc_input (src, "invalid chr specification");
	  }

	case '"':
	  /* Sym with strange chars.  */
	  stream_ungetc (src, c);
	  struct reader_object ro = read_next_ro (src);
	  RETURN_RO (pwg_str_to_sym (ro.obj), OBJ);

	case '(':
	  /* Ary.  */
	  stream_ungetc (src, c);
	  RETURN_RO (chr ('a'), SPECIAL);

	case '[':
	  /* Hash.  */
	  RETURN_RO (chr ('h'), SPECIAL);

	case '{':
	  exc_input (src, "trying to read output-only object");

	default:
	  exc_input (src, "invalid input after $");
	}
      break; /* Not reached.  */

    case '(': case ')': case '|': case '@': case '%':
      /* Special separators.  */
      RETURN_RO (c, SPECIAL);

    case '\\':
      RETURN_RO (PWG_INTERNAL_IMPROPER, OBJ);

    case ':':
      exc_input (src, "unexpected ':'");

    default:
      /* Sym, Settor, Sint and Float.  */
      {
	pwg_t r, s;

	stream_ungetc (src, c);
	s = read_into_str (src, is_symchr_or_asgn);
	if (unsint (pwg_str_size (s)) == 0)
	  exc_input (src, "invalid input character in stream.");

	r = pwg_str_to_sint (s);
	if (unbooled (r))
	  RETURN_RO (r, OBJ);
	r = pwg_str_to_float (s);
	if (unbooled (r))
	  RETURN_RO (r, OBJ);

	/* Either sym or settor. */
	pwg_t last = sint (unsint (pwg_str_size (s)) - 1);
	if (unchr (pwg_str_ref (s, last)) != ':')
	  RETURN_RO (pwg_str_to_sym (s), OBJ);
	else
	  RETURN_RO (pwg_settor_new
		     (pwg_str_to_sym
		      (pwg_str_slice (s, sint (0), last))), OBJ);
      }
    }
  RETURN_RO (PWG_EOF, OBJ);
#undef RETURN_RO
}

static struct reader_object saved;
static bool have_saved_p = false;

static void
save_ro (struct reader_object ro)
{
  if (PWG_GCC_EXPECT_FALSE (have_saved_p))
    err ("internal error: already safed a reader object");
  have_saved_p = true;
  saved = ro;
}

static struct reader_object
next_ro (stream_t src)
{
  if (have_saved_p)
    {
      have_saved_p = false;
      return saved;
    }
  return read_next_ro (src);
}


/* We have to collect lists at various places.  That's what a
   collector is for.  The car is the result and the CDR is the end. */
static pwg_t
collector_new (void)
{
  return pwg_cons (PWG_NIL, PWG_NIL);
}

static pwg_t
collector_result (pwg_t c)
{
  return pwg_car (c);
}

static pwg_t
collector_end (pwg_t c)
{
  return pwg_cdr (c);
}

static void
collector_add (pwg_t c, pwg_t o)
{
  pwg_t tmp = pwg_cons (o, PWG_NIL);

  if (!eq_p (pwg_cdr (c), PWG_NIL))
    pwg_set_cdr (pwg_cdr (c), tmp);
  else
    pwg_set_car (c, tmp);
  pwg_set_cdr (c, tmp);
}

/* Parser.  */
pwg_t
in_rec (stream_t src, int nesting)
{
  /* PWG_INTERNAL_LIST means that a list has been correctly closed,
     whether this is by EOF or by ')'. */
  pwg_t open_space_p = PWG_FALSE; /* Used for a clever trick below. */
  struct reader_object ro = { chr ('('), SPECIAL };
  if (nesting != 0)
    ro = next_ro (src);

  switch (ro.type)
    {
    case OBJ:
      return ro.obj;

    case SPECIAL:
      switch (unchr (ro.obj))
	{
	case ')':
	  if (nesting != 1)
	    return PWG_INTERNAL_ENDLIST;
	  else
	    exc_input (src, "parse error: unexpected ')'");

	case '(':
	  {
	    pwg_t collect = collector_new ();
	    while (1)
	      {
		pwg_t obj = in_rec (src, nesting + 1);
		if (eq_p (obj, PWG_INTERNAL_ENDLIST))
		  return collector_result (collect);

		/* Special case: improper list. We've just read the '\'. */
		if (eq_p (obj, PWG_INTERNAL_IMPROPER))
		  {
		    obj = in_rec (src, nesting + 1);
		    /* Must be normal object and next one must be ')'. */
		    if (eq_p (obj, PWG_INTERNAL_ENDLIST)
			|| eq_p (obj, PWG_INTERNAL_IMPROPER)
			|| !eq_p (in_rec (src, nesting + 1 ),
				  PWG_INTERNAL_ENDLIST))
		      exc_input (src,
				 "parse error: wrong ending of improper list");

		    if (eq_p (collector_end (collect), PWG_NIL))
		      return obj; /* (\ x) -> x */
		    pwg_set_cdr (collector_end (collect), obj);
		    return collector_result (collect);		      
		  }
		else /* Normal case: add an element to the list. */
		  collector_add (collect, obj);
	      }
	  }
	  break; /* Not reached */

	case '\'':
	  return pwg_quotor_new (in_rec (src, nesting + 1));

	read_pipe_proc:
	  open_space_p = PWG_TRUE;
	case '|':
	  {
	    pwg_t args = collector_new ();
	    while (1)
	      {
		pwg_t next = in_rec (src, nesting + 1);
		if (unbooled (pwg_cons_p (next)) || eq_p (next, PWG_NIL))
		  return pwg_proc_new (next, collector_result (args),
				       PWG_NIL, PWG_FALSE, open_space_p);

		if (!unbooled (pwg_sym_p (next)))
		  exc_input (src,
			     "parse error: argument list is invalid");
		collector_add (args, next);
	      }
	  }
	  break; /* Not reached */

	read_at_proc:
	  open_space_p = PWG_TRUE;
	case '@':
	  {
	    pwg_t args = collector_new ();
	    while (1)
	      {
		ro = next_ro (src);
		if (ro.type == OBJ && unbooled (pwg_sym_p (ro.obj)))
		  {
		    collector_add (args, ro.obj);
		    continue;
		  }
		if (ro.type == SPECIAL && unchr (ro.obj) == '(')
		  {
		    save_ro (ro);
		    return pwg_proc_new (in_rec (src, nesting + 1),
					 collector_result (args), PWG_NIL,
					 PWG_TRUE, open_space_p);
		  }
		if (ro.type == SPECIAL && unchr (ro.obj) == '|')
		  {
		    save_ro (ro);
		    pwg_t proc = in_rec (src, nesting);
		    return pwg_proc_new (pwg_proc_code (proc),
					 collector_result (args),
					 pwg_proc_args (proc),
					 PWG_TRUE, open_space_p);
		  }
		exc_input (src, "parse error: invalid token in argument list");
	      }
	  }
	  break; /* Not reached */

	case '%':
	  ro = next_ro (src);
	  if (ro.type == SPECIAL && unchr (ro.obj) == '(')
	    {
	      save_ro (ro);
	      return pwg_proc_new (in_rec (src, nesting + 1), PWG_NIL, PWG_NIL,
				   PWG_FALSE, PWG_TRUE);
	    }
	  if (ro.type == SPECIAL && unchr (ro.obj) == '@')
	    goto read_at_proc;
	  if (ro.type == OBJ && unbooled (pwg_sym_p (ro.obj)))
	    {
	      save_ro (ro);
	      goto read_pipe_proc;
	    }
	  exc_input (src, "parse error: invalid token after '%'");
	  break; /* Not reached */

	case 'a':
	  return pwg_ary_new (in_rec (src, nesting + 1));

	case 'h':
	  {
	    pwg_t has_def = in_rec (src, nesting + 1);
	    pwg_t result = pwg_hash_new (sint (16));
	    if (unbooled (has_def))
	      pwg_hash_set_default (result, in_rec (src, nesting + 1));

	    for (;;)
	      {
		pwg_t entry = in_rec (src, nesting + 1);

		if (eq_p (entry, sym ("]")))
		  return result;
		pwg_hash_set_item (result, pwg_car (entry), pwg_cdr (entry));
	      }
	  }

	default:
	  err ("internal parser error (should never happen)");
	}

    case XEOF:
      if (nesting != 1)
	exc_input (src, "parse error: eof in list");
      else
	/* This works because on start of file, we simulate a '('.
	 * See the beginning of this function. */
	return PWG_INTERNAL_ENDLIST;
    }
  abort (); /* Not reached. */
  return PWG_NONE;
}

pwg_t
in (stream_t src)
{
  return in_rec (src, 0);
}

pwg_t
pwg_stream_in (pwg_t stream)
{
  return in (stream_ref (stream));
}



void
setup_stream (void)
{
  stdout_stream.name = str_new ("<stdout>");
  stderr_stream.name = str_new ("<stderr>");
  stdin_stream.name = str_new ("<stdin>");

  stdout_hook.fp = stdout;
  stderr_hook.fp = stderr;
  stdin_hook.fp = stdin;

  the_stdout = pwgize_stream (stream_stdout ());
  the_stderr = pwgize_stream (stream_stderr ());
  the_stdin = pwgize_stream (stream_stdin ());
}

void
init_stream (void)
{
  def_core_psub ("stream-read", pwg_stream_read, 1);
  def_core_psub ("stream-ungetc", pwg_stream_ungetc, 2);
  def_core_psub ("stream-lookahead", pwg_stream_lookahead, 1);
  def_core_psub ("stream-write", pwg_stream_write, 2);
  def_core_psub ("stream-name", pwg_stream_name, 1);
  def_core_psub ("stream-input-line", pwg_stream_input_line, 1);
  def_core_psub ("stream-read-line", pwg_stream_read_line, 1);

  def_core_psub ("str-src-new", pwg_str_src_new, 1);
  def_core_psub ("str-dst-new", pwg_str_dst_new, 0);
  def_core_psub ("str-dst->str", pwg_str_dst_to_str, 1);

  def_core_psub ("file-src-new", pwg_file_src_new, 1);
  def_core_psub ("file-src-close", pwg_file_src_close, 1);
  def_core_psub ("file-dst-new", pwg_file_dst_new, 2);
  def_core_psub ("file-dst-close", pwg_file_dst_close, 1);

  def_core_psub ("stream-stdin", pwg_stream_stdin, 0);
  def_core_psub ("stream-stdout", pwg_stream_stdout, 0);
  def_core_psub ("stream-stderr", pwg_stream_stderr, 0);

  def_core_psub ("stream-in", pwg_stream_in, 1);
  def_core_psub ("stream-out", pwg_stream_out, 2);
}

void
finish_stream (void)
{}
