/* symkey.c -- Symbol implementation.
 * Copyright (C) 2014 Wolfgang Jaehrling
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

// FIXME: we don't have keywords yet. should we have them?

#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "symkey.h"
#include "mem.h"
#include "hash.h"
#include "compound.h"
#include "eval.h"

static hash_t sym_ht;

// FIXME: this is for debugging the GC only.
#include "stream.h"
void out (stream_t dest, pwg_t obj, natural_t remain);
static stream_t
stream_ref (pwg_t o)
{
  return *((stream_t *) mem_ref_obj (o, TYPE_STREAM));
}

static void
mark_sym_iter (void *h, natural_t k, natural_t v)
{
  /*if (!GC_MARKED_P (((pwg_type_t *)v)[-1])) { //FIXME
  out (stream_ref (pwg_stream_stdout ()),
       pwg_new (v),
       2);
       puts (" <- sym was not marked before"); }*/
  gc_mark (pwg_new (v));
}

void
mark_symtable (void)
{
  hash_iterate (sym_ht, mark_sym_iter, NULL);
}

/* Return the existing sym with text STR.  Afterwards, *KEY will be
   the key that enables you to find it in the hash.  This function can
   be used to test whether an entry exists in the table, because if
   the object does not exist yet, the `void' object will be returned,
   and *KEY will refer to a free slot in the hash (where the sym can
   be created).  */
static pwg_t
sym_query (pwg_t str, natural_t *key)
{
  *key = unsint (pwg_str_hash (str));
  pwg_t result = pwg_new (hash_val (sym_ht, *key));

  /* This code assumes that the hash table will be resized before it
     is filled entirely, so we will eventually get a null pointer if
     the entry does not exist yet.  */
  while (1)
    {
      /* The sym does not exist.  */
      if (pwg_get (result) == 0)
	return PWG_VOID;

      /* We found it.   */
      if (unbooled (pwg_str_sym_eql_p (str, result)))
	return result;

      /* Try again.  */
      (*key)++;
      result = pwg_new (hash_val (sym_ht, *key));
    }
  /* Not reached.  */
  return PWG_VOID;
}



pwg_t
pwg_str_to_sym (pwg_t str)
{
  natural_t key;
  pwg_t result;

  result = sym_query (str, &key);
  if (eq_p (result, PWG_VOID))
    {
      int exc;
      /* No entry yet, so we create it.  */
      EXC_TRY
	result = pwg_str_copy (str);
      EXC_ENSURE (exc)
      {} //FIXME: cn wae get rid of that?
      EXC_FWD (exc);
      mem_set_type (result, TYPE_SYM);
      hash_set (sym_ht, key, pwg_get (result));
    }
  return result;
}

pwg_t
pwg_sym_to_str (pwg_t sym)
{
  mem_ref_obj (sym, TYPE_SYM); /* Typecheck.  */
  mem_set_type (sym, TYPE_STR);
  pwg_t result = pwg_str_copy (sym);
  mem_set_type (sym, TYPE_SYM);
  return result;
}

DEF_PRED (pwg_sym_p, o, mem_is_type_p (o, TYPE_SYM));

pwg_t
sym (const char *str)
{
  return pwg_str_to_sym (str_new (str));
}



void
setup_symkey (void)
{
  sym_ht = hash_new (1024);
}

void
init_symkey (void)
{
  def_core_psub ("str->sym", pwg_str_to_sym, 1);
  def_core_psub ("sym->str", pwg_sym_to_str, 1);
  def_core_psub ("sym?", pwg_sym_p, 1);
}

void
finish_symkey (void)
{
  hash_free (sym_ht);
}
